# Unit test DateTimeFeatureTransformer
# ==============================================================================
import pytest
import re
import pandas as pd
import numpy as np
from skforecast.preprocessing import DateTimeFeatureTransformer

# Fixtures
from .fixtures_preprocessing import features_all_onehot

if pd.__version__ < '2.2.0':
    freq_h = "H"
else:
    freq_h = "h"


def test_create_datetime_features_invalid_input_type():
    """
    Test that DateTimeFeatureTransformer raises a ValueError when input is not 
    a pandas DataFrame or Series.
    """
    with pytest.raises(TypeError, match="Input `X` must be a pandas Series or DataFrame"):
        DateTimeFeatureTransformer().fit_transform([1, 2, 3])


def test_create_datetime_features_no_datetime_index():
    """
    Test that DateTimeFeatureTransformer raises a ValueError when input does not 
    have a pandas DatetimeIndex index.
    """
    df = pd.DataFrame({"a": [1, 2, 3]})
    with pytest.raises(TypeError, match="Input `X` must have a pandas DatetimeIndex"):
        DateTimeFeatureTransformer().fit_transform(df)


def test_create_datetime_features_invalid_encoding():
    """
    Test that DateTimeFeatureTransformer raises a ValueError when encoding is not 
    one of 'cyclical', 'onehot' or None.
    """
    df = pd.DataFrame(
        np.random.rand(5, 3),
        columns=["col_1", "col_2", "col_3"],
        index=pd.date_range(start="1/1/2022", end="1/5/2022", freq="D"),
    )

    with pytest.raises(
        ValueError, match="Encoding must be one of 'cyclical', 'onehot' or None"
    ):
        DateTimeFeatureTransformer(encoding="invalid encoding").fit_transform(df)


def test_create_datetime_features_invalid_feature_name():
    """
    Test that DateTimeFeatureTransformer raises a ValueError when a feature name is
    not valid.
    """
    df = pd.DataFrame(
        np.random.rand(5, 3),
        columns=["col_1", "col_2", "col_3"],
        index=pd.date_range(start="1/1/2022", end="1/5/2022", freq="D"),
    )

    err_msg = re.escape(
        "Features {'invalid_feature'} are not supported. Supported features are "
        "['year', 'month', 'week', 'day_of_week', 'day_of_year', 'day_of_month', "
        "'weekend', 'hour', 'minute', 'second']."
    )
    with pytest.raises(ValueError, match=err_msg):
        DateTimeFeatureTransformer(features=["invalid_feature"]).fit_transform(df)


def test_create_datetime_features_output_columns_when_cyclical_encoding():
    """
    Test that DateTimeFeatureTransformer returns the expected columns when encoding is 'cyclical'.
    """
    df = pd.DataFrame(
        np.random.rand(5, 3),
        columns=["col_1", "col_2", "col_3"],
        index=pd.date_range(start="1/1/2022", end="1/5/2022", freq="D"),
    )

    results = DateTimeFeatureTransformer(encoding="cyclical").fit_transform(df)
    expected_features = [
        "year",
        "weekend",
        "month_sin",
        "month_cos",
        "week_sin",
        "week_cos",
        "day_of_week_sin",
        "day_of_week_cos",
        "day_of_month_sin",
        "day_of_month_cos",
        "day_of_year_sin",
        "day_of_year_cos",
        "hour_sin",
        "hour_cos",
        "minute_sin",
        "minute_cos",
        "second_sin",
        "second_cos",
    ]

    assert all([feature in results.columns for feature in expected_features])
    assert len(results) == len(df)


def test_create_datetime_features_output_columns_when_onehot_encoding():
    """
    Test that DateTimeFeatureTransformer returns the expected columns when encoding is 'onehot'.
    """
    index = pd.date_range(start="2021-01-01", end="2023-01-01", freq=freq_h)
    df = pd.DataFrame(
        np.random.rand(len(index), 3),
        columns=["col_1", "col_2", "col_3"],
        index=index,
    )

    results = DateTimeFeatureTransformer(encoding="onehot").fit_transform(df)

    assert all([feature in features_all_onehot for feature in results.columns])
    assert len(results) == len(df)


def test_create_datetime_features_output_columns_when_None_encoding():
    """
    Test that DateTimeFeatureTransformer returns the expected columns when encoding is 'None'.
    """
    index = pd.date_range(start="2021-01-01", end="2023-01-01", freq=freq_h)
    df = pd.DataFrame(
        np.random.rand(len(index), 3),
        columns=["col_1", "col_2", "col_3"],
        index=index,
    )
    results = DateTimeFeatureTransformer(encoding=None).fit_transform(df)
    expected_features = [
        "year",
        "month",
        "week",
        "day_of_week",
        "day_of_month",
        "day_of_year",
        "weekend",
        "hour",
        "minute",
        "second",
    ]
    assert all(results.columns == expected_features)
    assert len(results) == len(df)


def test_create_datetime_features_output_when_features_year_month_encoding_cyclical():
    """
    Test that DateTimeFeatureTransformer returns the expected columns when features
     is ['year', 'month'] and encoding is 'cyclical'.
    """
    df = pd.DataFrame(
        np.random.rand(5, 3),
        columns=["col_1", "col_2", "col_3"],
        index=pd.date_range(start="1/1/2022", end="1/5/2022", freq="D"),
    )

    results = DateTimeFeatureTransformer(
        features=["year", "month", "weekend"], encoding="cyclical"
    ).fit_transform(df)
    expected = pd.DataFrame(
        {
            "year": {
                pd.Timestamp("2022-01-01 00:00:00"): 2022,
                pd.Timestamp("2022-01-02 00:00:00"): 2022,
                pd.Timestamp("2022-01-03 00:00:00"): 2022,
                pd.Timestamp("2022-01-04 00:00:00"): 2022,
                pd.Timestamp("2022-01-05 00:00:00"): 2022,
            },
            "weekend": {
                pd.Timestamp("2022-01-01 00:00:00"): 1,
                pd.Timestamp("2022-01-02 00:00:00"): 1,
                pd.Timestamp("2022-01-03 00:00:00"): 0,
                pd.Timestamp("2022-01-04 00:00:00"): 0,
                pd.Timestamp("2022-01-05 00:00:00"): 0,
            },
            "month_sin": {
                pd.Timestamp("2022-01-01 00:00:00"): 0.49999999999999994,
                pd.Timestamp("2022-01-02 00:00:00"): 0.49999999999999994,
                pd.Timestamp("2022-01-03 00:00:00"): 0.49999999999999994,
                pd.Timestamp("2022-01-04 00:00:00"): 0.49999999999999994,
                pd.Timestamp("2022-01-05 00:00:00"): 0.49999999999999994,
            },
            "month_cos": {
                pd.Timestamp("2022-01-01 00:00:00"): 0.8660254037844387,
                pd.Timestamp("2022-01-02 00:00:00"): 0.8660254037844387,
                pd.Timestamp("2022-01-03 00:00:00"): 0.8660254037844387,
                pd.Timestamp("2022-01-04 00:00:00"): 0.8660254037844387,
                pd.Timestamp("2022-01-05 00:00:00"): 0.8660254037844387,
            },
        }
    ).asfreq("D").astype(
        {'year': int, 'weekend': int}
    )

    pd.testing.assert_frame_equal(results, expected)


def test_create_datetime_features_output_when_features_year_month_encoding_onehot():
    """
    Test that DateTimeFeatureTransformer returns the expected columns when features
     is ['year', 'month'] and encoding is 'onehot'.
    """
    df = pd.DataFrame(
        np.random.rand(5, 3),
        columns=["col_1", "col_2", "col_3"],
        index=pd.date_range(start="1/1/2022", end="1/5/2022", freq="D"),
    )

    results = DateTimeFeatureTransformer(
        features=["year", "month", "weekend"], encoding="onehot"
    ).fit_transform(
        df,
    )
    expected = pd.DataFrame(
        {
            "year_2022": {
                pd.Timestamp("2022-01-01 00:00:00"): 1,
                pd.Timestamp("2022-01-02 00:00:00"): 1,
                pd.Timestamp("2022-01-03 00:00:00"): 1,
                pd.Timestamp("2022-01-04 00:00:00"): 1,
                pd.Timestamp("2022-01-05 00:00:00"): 1,
            },
            "month_1": {
                pd.Timestamp("2022-01-01 00:00:00"): 1,
                pd.Timestamp("2022-01-02 00:00:00"): 1,
                pd.Timestamp("2022-01-03 00:00:00"): 1,
                pd.Timestamp("2022-01-04 00:00:00"): 1,
                pd.Timestamp("2022-01-05 00:00:00"): 1,
            },
            "weekend_0": {
                pd.Timestamp("2022-01-01 00:00:00"): 0,
                pd.Timestamp("2022-01-02 00:00:00"): 0,
                pd.Timestamp("2022-01-03 00:00:00"): 1,
                pd.Timestamp("2022-01-04 00:00:00"): 1,
                pd.Timestamp("2022-01-05 00:00:00"): 1,
            },
            "weekend_1": {
                pd.Timestamp("2022-01-01 00:00:00"): 1,
                pd.Timestamp("2022-01-02 00:00:00"): 1,
                pd.Timestamp("2022-01-03 00:00:00"): 0,
                pd.Timestamp("2022-01-04 00:00:00"): 0,
                pd.Timestamp("2022-01-05 00:00:00"): 0,
            },
        }
    ).asfreq("D").astype(
        {'year_2022': int, 'month_1': int, 'weekend_0': int, 'weekend_1': int}
    )

    pd.testing.assert_frame_equal(results, expected)


def test_create_datetime_features_output_when_features_year_month_encoding_None():
    """
    Test that DateTimeFeatureTransformer returns the expected columns when features
     is ['year', 'month'] and encoding is None.
    """
    df = pd.DataFrame(
        np.random.rand(5, 3),
        columns=["col_1", "col_2", "col_3"],
        index=pd.date_range(start="1/1/2022", end="1/5/2022", freq="D"),
    )

    results = DateTimeFeatureTransformer(
        features=["year", "month", "weekend"], encoding=None
    ).fit_transform(
        df,
    )
    expected = pd.DataFrame(
        {
            "year": {
                pd.Timestamp("2022-01-01 00:00:00"): 2022,
                pd.Timestamp("2022-01-02 00:00:00"): 2022,
                pd.Timestamp("2022-01-03 00:00:00"): 2022,
                pd.Timestamp("2022-01-04 00:00:00"): 2022,
                pd.Timestamp("2022-01-05 00:00:00"): 2022,
            },
            "month": {
                pd.Timestamp("2022-01-01 00:00:00"): 1,
                pd.Timestamp("2022-01-02 00:00:00"): 1,
                pd.Timestamp("2022-01-03 00:00:00"): 1,
                pd.Timestamp("2022-01-04 00:00:00"): 1,
                pd.Timestamp("2022-01-05 00:00:00"): 1,
            },
            "weekend": {
                pd.Timestamp("2022-01-01 00:00:00"): 1,
                pd.Timestamp("2022-01-02 00:00:00"): 1,
                pd.Timestamp("2022-01-03 00:00:00"): 0,
                pd.Timestamp("2022-01-04 00:00:00"): 0,
                pd.Timestamp("2022-01-05 00:00:00"): 0,
            },
        }
    ).asfreq("D").astype(
        {'year': int, 'month': int, 'weekend': int}
    )

    pd.testing.assert_frame_equal(results, expected)


def test_create_datetime_features_output_when_features_year_month_encoding_cyclical_and_custom_max_values():
    """
    Test that DateTimeFeatureTransformer returns the expected columns when features
    is ['year', 'month'] and encoding is 'cyclical' with custom max values.
    """

    df = pd.DataFrame(
        np.random.rand(5, 3),
        columns=["col_1", "col_2", "col_3"],
        index=pd.DatetimeIndex(
            ["2022-01-31", "2022-02-28", "2022-03-31", "2022-04-30", "2022-05-31"]
        ),
    )

    results = DateTimeFeatureTransformer(
        features=["year", "month", "weekend"],
        encoding="cyclical",
        max_values={"month": 6},
    ).fit_transform(df)

    expected = pd.DataFrame(
        {
            "year": {
                pd.Timestamp("2022-01-31 00:00:00"): 2022,
                pd.Timestamp("2022-02-28 00:00:00"): 2022,
                pd.Timestamp("2022-03-31 00:00:00"): 2022,
                pd.Timestamp("2022-04-30 00:00:00"): 2022,
                pd.Timestamp("2022-05-31 00:00:00"): 2022,
            },
            "weekend": {
                pd.Timestamp("2022-01-31 00:00:00"): 0,
                pd.Timestamp("2022-02-28 00:00:00"): 0,
                pd.Timestamp("2022-03-31 00:00:00"): 0,
                pd.Timestamp("2022-04-30 00:00:00"): 1,
                pd.Timestamp("2022-05-31 00:00:00"): 0,
            },
            "month_sin": {
                pd.Timestamp("2022-01-31 00:00:00"): 0.8660254037844386,
                pd.Timestamp("2022-02-28 00:00:00"): 0.8660254037844387,
                pd.Timestamp("2022-03-31 00:00:00"): 1.2246467991473532e-16,
                pd.Timestamp("2022-04-30 00:00:00"): -0.8660254037844384,
                pd.Timestamp("2022-05-31 00:00:00"): -0.8660254037844386,
            },
            "month_cos": {
                pd.Timestamp("2022-01-31 00:00:00"): 0.5000000000000001,
                pd.Timestamp("2022-02-28 00:00:00"): -0.4999999999999998,
                pd.Timestamp("2022-03-31 00:00:00"): -1.0,
                pd.Timestamp("2022-04-30 00:00:00"): -0.5000000000000004,
                pd.Timestamp("2022-05-31 00:00:00"): 0.5000000000000001,
            },
        }
    ).astype(
        {'year': int, 'weekend': int}
    )

    pd.testing.assert_frame_equal(results, expected)



# Unit test root_mean_squared_scaled_error
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from skforecast.metrics import root_mean_squared_scaled_error


def test_root_mean_squared_scaled_error_input_types():
    """
    Test input types of root_mean_squared_scaled_error.
    """
    y_true = np.array([1.0, 2.0, 3.0, 4.0])
    y_pred = np.array([1.1, 1.9, 3.2, 3.8])
    y_train = np.array([0.5, 1.5, 2.5, 3.5, 4.5])

    err_msg = re.escape("`y_true` must be a pandas Series or numpy ndarray")
    with pytest.raises(TypeError, match = err_msg):
        root_mean_squared_scaled_error([1, 2, 3], y_pred, y_train)
    
    err_msg = re.escape("`y_pred` must be a pandas Series or numpy ndarray")
    with pytest.raises(TypeError, match = err_msg):
        root_mean_squared_scaled_error(y_true, [1, 2, 3], y_train)
    
    err_msg = re.escape("`y_train` must be a list, pandas Series or numpy ndarray")
    with pytest.raises(TypeError, match = err_msg):
        root_mean_squared_scaled_error(y_true, y_pred, 'not_valid_input')
    
    err_msg = re.escape(
        ("When `y_train` is a list, each element must be a pandas Series "
         "or numpy ndarray")
    )
    with pytest.raises(TypeError, match = err_msg):
        root_mean_squared_scaled_error(y_true, y_pred, [1, 2, 3])


def test_root_mean_squared_scaled_error_input_length():
    """
    Test input lengths of root_mean_squared_scaled_error.
    """
    y_true = np.array([1.0, 2.0, 3.0, 4.0])
    y_pred = np.array([1.1, 1.9, 3.2])
    y_train = np.array([0.5, 1.5, 2.5, 3.5, 4.5])

    err_msg = re.escape("`y_true` and `y_pred` must have the same length")
    with pytest.raises(ValueError, match = err_msg):
        root_mean_squared_scaled_error(y_true, y_pred, y_train)

    err_msg = re.escape("`y_true` and `y_pred` must have the same length")
    with pytest.raises(ValueError, match = err_msg):
        root_mean_squared_scaled_error(y_true, y_pred, y_train)


def test_root_mean_squared_scaled_error_empty_input():
    """
    Test empty input of root_mean_squared_scaled_error.
    """
    y_true = np.array([1.0, 2.0, 3.0, 4.0])
    y_pred = np.array([1.1, 1.9, 3.2, 3.8])
    y_train = np.array([0.5, 1.5, 2.5, 3.5, 4.5])

    err_msg = re.escape("`y_true` and `y_pred` must have at least one element")
    with pytest.raises(ValueError, match = err_msg):
        root_mean_squared_scaled_error(np.array([]), np.array([]), y_train)


def test_root_mean_squared_scaled_error_output():
    """
    Test input types of root_mean_squared_scaled_error.
    """
    y_true = np.array([1.0, 2.0, 3.0, 4.0])
    y_pred = np.array([1.1, 1.9, 3.2, 3.8])
    y_train = np.array([0.5, 1.5, 2.5, 3.5, 4.5])

    expected_rmsse = np.sqrt(np.mean((y_true - y_pred) ** 2)) / np.sqrt(np.mean(np.diff(y_train) ** 2))
    
    assert np.isclose(root_mean_squared_scaled_error(y_true, y_pred, y_train), expected_rmsse)
    

def test_pandas_series_input():
    """
    Test pandas Series input of root_mean_squared_scaled_error.
    """
    y_true = np.array([1.0, 2.0, 3.0, 4.0])
    y_pred = np.array([1.1, 1.9, 3.2, 3.8])
    y_train = np.array([0.5, 1.5, 2.5, 3.5, 4.5])
    y_true_series = pd.Series(y_true)
    y_pred_series = pd.Series(y_pred)
    y_train_series = pd.Series(y_train)
    
    expected_rmsse = np.sqrt(np.mean((y_true - y_pred) ** 2)) / np.sqrt(np.mean(np.diff(y_train) ** 2))
    
    assert np.isclose(root_mean_squared_scaled_error(y_true_series, y_pred_series, y_train_series), expected_rmsse)


# Unit test predict ForecasterStats
# ==============================================================================
import re
import pytest
import platform
import pandas as pd
from sklearn.exceptions import NotFittedError
from skforecast.stats import Sarimax, Arar
from skforecast.recursive import ForecasterStats
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from aeon.forecasting.stats import ARIMA

# Fixtures
from .fixtures_forecaster_stats import y
from .fixtures_forecaster_stats import y_lw
from .fixtures_forecaster_stats import exog
from .fixtures_forecaster_stats import exog_lw
from .fixtures_forecaster_stats import exog_predict
from .fixtures_forecaster_stats import exog_lw_predict
from .fixtures_forecaster_stats import y_datetime
from .fixtures_forecaster_stats import y_lw_datetime
from .fixtures_forecaster_stats import exog_datetime
from .fixtures_forecaster_stats import exog_lw_datetime
from .fixtures_forecaster_stats import exog_predict_datetime
from .fixtures_forecaster_stats import exog_lw_predict_datetime
from .fixtures_forecaster_stats import df_exog
from .fixtures_forecaster_stats import df_exog_lw
from .fixtures_forecaster_stats import df_exog_predict
from .fixtures_forecaster_stats import df_exog_lw_predict
from .fixtures_forecaster_stats import df_exog_datetime
from .fixtures_forecaster_stats import df_exog_lw_datetime
from .fixtures_forecaster_stats import df_exog_predict_datetime
from .fixtures_forecaster_stats import df_exog_lw_predict_datetime


def test_predict_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterStats(estimator=Sarimax(order=(1, 1, 1)))

    err_msg = re.escape(
        ("This Forecaster instance is not fitted yet. Call `fit` with "
         "appropriate arguments before using predict.")
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.predict(steps=5)


def test_predict_ValueError_when_ForecasterStats_last_window_exog_is_not_None_and_last_window_is_not_provided():
    """
    Check ValueError is raised when last_window_exog is not None, but 
    last_window is not provided.
    """
    forecaster = ForecasterStats(estimator=Sarimax(order=(1, 1, 1)))
    forecaster.fit(y=y, exog=exog)
    
    err_msg = re.escape(
        ("To make predictions unrelated to the original data, both "
         "`last_window` and `last_window_exog` must be provided.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict(
            steps            = 5, 
            exog             = exog_predict, 
            last_window      = None, 
            last_window_exog = exog
        )


def test_predict_ValueError_when_ForecasterStats_last_window_exog_is_None_and_included_exog_is_true():
    """
    Check ValueError is raised when last_window_exog is None, but included_exog
    is True and last_window is provided.
    """
    forecaster = ForecasterStats(estimator=Sarimax(order=(1, 1, 1)))
    forecaster.fit(y=y, exog=exog)
    
    err_msg = re.escape(
        ("Forecaster trained with exogenous variable/s. To make predictions "
         "unrelated to the original data, same variable/s must be provided "
         "using `last_window_exog`.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict(
            steps            = 5, 
            exog             = exog_lw_predict, 
            last_window      = y_lw, 
            last_window_exog = None
        )


@pytest.mark.parametrize("kwargs, data", 
                         [({'order': (1, 0, 1), 
                            'seasonal_order': (0, 0, 0, 0)}, 
                           {'win': [0.63432268, 0.62507372, 0.61595962, 0.60697841, 0.59812815],
                            'linux': [0.60535333, 0.59654171, 0.58785836, 0.5793014 , 0.570869]}), 
                          ({'order': (1, 1, 1), 
                            'seasonal_order': (1, 1, 1, 2)}, 
                           {'win': [0.5366165, 0.55819701, 0.49539926, 0.51944837, 0.45417575],
                            'linux': [0.5366165 , 0.55819701, 0.49539926, 0.51944837, 0.45417575]})])
def test_predict_output_ForecasterStats_skforecast_Sarimax(kwargs, data):
    """
    Test predict output of ForecasterStats using Sarimax from skforecast.
    """
    system = "win" if platform.system() == "Windows" else 'linux'

    forecaster = ForecasterStats(
                     estimator = Sarimax(maxiter=1000, method='cg', disp=False, **kwargs)
                 )
    forecaster.fit(y=y)
    predictions = forecaster.predict(steps=5)
    expected = pd.Series(
                   data  = data[system],
                   index = pd.RangeIndex(start=50, stop=55, step=1),
                   name  = 'pred'
               )

    pd.testing.assert_series_equal(predictions, expected)


@pytest.mark.parametrize("kwargs, data", 
                         [({'order': (1, 0, 1), 
                            'seasonal_order': (0, 0, 0, 0)}, 
                            [0.59929905, 0.61299725, 0.6287311 , 0.64413557, 0.66195978]), 
                          ({'order': (1, 1, 1), 
                            'seasonal_order': (1, 1, 1, 2)}, 
                            [0.47217517, 0.57747478, 0.58655865, 0.69219403, 0.71031467])])
def test_predict_output_ForecasterStats_with_exog(kwargs, data):
    """
    Test predict output of ForecasterStats with exogenous variables.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(maxiter=1000, method='cg', disp=False, **kwargs)
                 )
    forecaster.fit(y=y, exog=exog)
    predictions = forecaster.predict(steps=5, exog=exog_predict)
    expected = pd.Series(
                   data  = data,
                   index = pd.RangeIndex(start=50, stop=55, step=1),
                   name  = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


@pytest.mark.parametrize("kwargs, data", 
                         [({'order': (1, 0, 1), 
                            'seasonal_order': (0, 0, 0, 0)}, 
                            {'win': [0.60290703, 0.60568721, 0.60451413, 0.6050091, 0.60480025],
                            'linux': [0.60290703, 0.60568721, 0.60451413, 0.6050091 , 0.60480025]}
                        )])
def test_predict_output_ForecasterStats_with_transform_y(kwargs, data):
    """
    Test predict output of ForecasterStats with a StandardScaler() as transformer_y.
    """
    system = "win" if platform.system() == "Windows" else 'linux'
        
    forecaster = ForecasterStats(
                     estimator     = Sarimax(maxiter=1000, method='cg', disp=False, **kwargs),
                     transformer_y = StandardScaler()
                 )
    forecaster.fit(y=y)
    predictions = forecaster.predict(steps=5)
    expected = pd.Series(
                   data  = data[system],
                   index = pd.RangeIndex(start=50, stop=55, step=1),
                   name  = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


@pytest.mark.parametrize("kwargs, data", 
                         [({'order': (1, 0, 1), 
                            'seasonal_order': (0, 0, 0, 0)},
                            {'win': [0.60687311, 0.62484493, 0.63515416, 0.67730912, 0.69458838],
                            'linux': [0.60687186, 0.62484336, 0.63515295, 0.67730812, 0.69458769]}
                            )])
def test_predict_output_ForecasterStats_with_transform_y_and_transform_exog(kwargs, data):
    """
    Test predict output of ForecasterStats, StandardScaler
    as transformer_y and transformer_exog as transformer_exog.
    """
    system = "win" if platform.system() == "Windows" else 'linux'
    
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )

    forecaster = ForecasterStats(
                     estimator        = Sarimax(maxiter=1000, method='cg', disp=False, **kwargs),
                     transformer_y    = StandardScaler(),
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y, exog=df_exog)
    predictions = forecaster.predict(steps=5, exog=df_exog_predict)
    expected = pd.Series(
                   data  = data[system],
                   index = pd.RangeIndex(start=50, stop=55, step=1),
                   name  = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


def test_predict_ValueError_when_last_window_index_does_not_follow_training_set():
    """
    Raise ValueError if `last_window` index does not start at the end 
    of the index seen by the forecaster.
    """
    y_test = pd.Series(data=y_datetime.to_numpy())
    y_test.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    lw_test = pd.Series(data=y_lw_datetime.to_numpy())
    lw_test.index = pd.date_range(start='2022-03-01', periods=50, freq='D')

    forecaster = ForecasterStats(estimator=Sarimax(order=(1, 0, 1)))
    forecaster.fit(y=y_test)

    err_msg = re.escape(
        (f"To make predictions unrelated to the original data, `last_window` "
         f"has to start at the end of the index seen by the forecaster.\n"
         f"    Series last index         : 2022-02-19 00:00:00.\n"
         f"    Expected index            : 2022-02-20 00:00:00.\n"
         f"    `last_window` index start : 2022-03-01 00:00:00.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict(steps=5, last_window=lw_test)


def test_predict_ValueError_when_last_window_exog_index_does_not_follow_training_set():
    """
    Raise ValueError if `last_window_exog` index does not start at the end 
    of the index seen by the forecaster.
    """
    y_test = pd.Series(data=y_datetime.to_numpy())
    y_test.index = pd.date_range(start='2022-01-01', periods=50, freq='D')
    lw_test = pd.Series(data=y_lw_datetime.to_numpy())
    lw_test.index = pd.date_range(start='2022-02-20', periods=50, freq='D')
    
    exog_test = pd.Series(data=exog_datetime.to_numpy(), name='exog')
    exog_test.index = pd.date_range(start='2022-01-01', periods=50, freq='D', name='exog')
    exog_pred_test = pd.Series(data=exog_predict_datetime.to_numpy(), name='exog')
    exog_pred_test.index = pd.date_range(start='2022-04-11', periods=10, freq='D', name='exog')
    lw_exog_test = pd.Series(data=exog_lw_datetime.to_numpy(), name='exog')
    lw_exog_test.index = pd.date_range(start='2022-03-01', periods=50, freq='D', name='exog')

    forecaster = ForecasterStats(estimator=Sarimax(order=(1, 0, 1)))
    forecaster.fit(y=y_test, exog=exog_test)

    err_msg = re.escape(
        (f"To make predictions unrelated to the original data, `last_window_exog` "
         f"has to start at the end of the index seen by the forecaster.\n"
         f"    Series last index              : 2022-02-19 00:00:00.\n"
         f"    Expected index                 : 2022-02-20 00:00:00.\n"
         f"    `last_window_exog` index start : 2022-03-01 00:00:00.")
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.predict(
            steps            = 5, 
            exog             = exog_pred_test, 
            last_window      = lw_test,
            last_window_exog = lw_exog_test
        )


@pytest.mark.parametrize("kwargs, data", 
                         [({'order': (1, 0, 1), 
                            'seasonal_order': (0, 0, 0, 0)},
                            {'win': [0.89534355, 0.88228868, 0.86942417, 0.85674723, 0.84425513],
                            'linux': [0.84149149, 0.82924261, 0.81717202, 0.80527714, 0.7935554]}),
                          ({'order': (1, 1, 1), 
                            'seasonal_order': (1, 1, 1, 2)},
                            {'win': [0.8079287 , 0.8570154 , 0.82680235, 0.88869186, 0.85565138],
                            'linux': [0.8079287 , 0.8570154 , 0.82680235, 0.88869186, 0.85565138]}
                            )])
def test_predict_output_ForecasterStats_with_last_window(kwargs, data):
    """
    Test predict output of ForecasterStats with `last_window`.
    """
    system = "win" if platform.system() == "Windows" else 'linux'
    
    forecaster = ForecasterStats(
                     estimator = Sarimax(maxiter=1000, method='cg', disp=False, **kwargs)
                 )
    forecaster.fit(y=y_datetime)
    predictions = forecaster.predict(steps=5, last_window=y_lw_datetime)
    expected = pd.Series(
                   data  = data[system],
                   index = pd.date_range(start='2100', periods=5, freq='YE'),
                   name  = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


@pytest.mark.parametrize("kwargs, data", 
                         [({'order': (1, 0, 1), 
                            'seasonal_order': (0, 0, 0, 0)}, 
                            [0.70551392, 0.68955739, 0.72525104, 0.74304522, 0.75825646]), 
                          ({'order': (1, 1, 1), 
                            'seasonal_order': (1, 1, 1, 2)}, 
                            [0.94305077, 1.06784624, 1.17677373, 1.30237101, 1.37057143])])
def test_predict_output_ForecasterStats_with_last_window_and_exog(kwargs, data):
    """
    Test predict output of ForecasterStats with exogenous variables and `last_window`.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(maxiter=1000, method='cg', disp=False, **kwargs)
                 )
    forecaster.fit(y=y_datetime, exog=exog_datetime)
    predictions = forecaster.predict(
                      steps            = 5, 
                      exog             = exog_lw_predict_datetime, 
                      last_window      = y_lw_datetime, 
                      last_window_exog = exog_lw_datetime
                  )

    expected = pd.Series(
                   data  = data,
                   index = pd.date_range(start='2100', periods=5, freq='YE'),
                   name  = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected, atol=0.0005)


@pytest.mark.parametrize("kwargs, data", 
                         [({'order': (1, 0, 1), 
                            'seasonal_order': (0, 0, 0, 0)}, 
                            [0.81663903, 0.77783205, 0.80523981, 0.85467197, 0.86644466])
                        ])
def test_predict_output_ForecasterStats_with_last_window_and_exog_and_transformers(kwargs, data):
    """
    Test predict output of ForecasterStats with exogenous variables, `last_window`
    and transformers.
    """
    transformer_exog = ColumnTransformer(
                           [('scale', StandardScaler(), ['exog_1']),
                            ('onehot', OneHotEncoder(), ['exog_2'])],
                           remainder = 'passthrough',
                           verbose_feature_names_out = False
                       )

    forecaster = ForecasterStats(
                     estimator        = Sarimax(maxiter=1000, method='cg', disp=False, **kwargs), 
                     transformer_y    = StandardScaler(),
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y_datetime, exog=df_exog_datetime)
    predictions = forecaster.predict(
                      steps            = 5, 
                      exog             = df_exog_lw_predict_datetime, 
                      last_window      = y_lw_datetime, 
                      last_window_exog = df_exog_lw_datetime
                  )

    expected = pd.Series(
                   data  = data,
                   index = pd.date_range(start='2100', periods=5, freq='YE'),
                   name  = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected, atol=0.0001)


@pytest.mark.parametrize("y          , idx", 
                         [(y         , pd.RangeIndex(start=0, stop=50)), 
                          (y_datetime, pd.date_range(start='2000', periods=50, freq='YE'))], 
                         ids = lambda values: f'y, index: {type(values)}')
def test_predict_ForecasterStats_updates_extended_index_twice(y, idx):
    """
    Test forecaster.extended_index_ is updated when using predict twice.
    """
    y_fit = y.iloc[:30].copy()

    forecaster = ForecasterStats(estimator=Sarimax(order=(1, 0, 0)))
    forecaster.fit(y=y_fit)

    lw_1 = y.iloc[30:40].copy()
    forecaster.predict(steps=5, last_window=lw_1)
    result_1 = forecaster.extended_index_.copy()
    expected_1 = idx[:40]

    lw_2 = y.iloc[40:].copy()
    forecaster.predict(steps=5, last_window=lw_2)

    pd.testing.assert_index_equal(result_1, expected_1)
    pd.testing.assert_index_equal(forecaster.extended_index_, idx)


def test_predict_output_ForecasterStats_with_Arar_estimator(y=y):
    """
    Test output of predict when using Arar as estimator in ForecasterStats
    """
    y = y.copy()
    y.index = pd.date_range(start="2000-01-01", periods=len(y), freq="D")
    estimator = Arar(max_ar_depth=26, max_lag=40)
    forecaster = ForecasterStats(estimator=estimator)
    forecaster.fit(y=y)
    predictions = forecaster.predict(steps=10)
    print(predictions)

    expected_results = pd.Series(
        data=[
            0.6545169424155248,
            0.693692742846921,
            0.8018875044998569,
            0.8215732635739413,
            0.8786870196348286,
            0.8879849576779746,
            1.0173957185481672,
            1.022217168929162,
            0.5688093026635382,
            0.6336566288056642,
        ],
        name="pred",
        index=pd.date_range(start="2000-02-20", periods=10, freq="D"),
    )

    pd.testing.assert_series_equal(predictions, expected_results)


def test_predict_output_ForecasterStats_with_aeon_ARIMA_estimator(y=y):
    """
    Test output of predict when using ARIMA from aeon as estimator in ForecasterStats
    """
    y = y.copy()
    y.index = pd.date_range(start="2000-01-01", periods=len(y), freq="D")
    estimator = ARIMA(p=4, d=1, q=1)
    forecaster = ForecasterStats(estimator=estimator)
    forecaster.fit(y=y)
    predictions = forecaster.predict(steps=10)
    print(predictions)

    expected_results = pd.Series(
        data=[
            0.68329597,
            0.7153698,
            0.72118068,
            0.71875056,
            0.70465789,
            0.68932917,
            0.68087982,
            0.67748653,
            0.67828572,
            0.68193945,
        ],
        name="pred",
        index=pd.date_range(start="2000-02-20", periods=10, freq="D"),
    )

    pd.testing.assert_series_equal(predictions, expected_results)


# Unit test init
# ==============================================================================
import pytest
import re
from ....drift_detection import PopulationDriftDetector


def test_init_exception_when_chunk_size_not_valid():
    """
    Test that an exception is raised when an invalid chunk_size is provided.
    """

    error_msg = re.escape(
        "`chunk_size` must be a positive integer, a string compatible with "
        "pandas DateOffset (e.g., 'D', 'W', 'M'), a pandas DateOffset object, or None."
    )

    chunk_size = ['a']
    with pytest.raises(TypeError, match=f"{error_msg} Got {type(chunk_size)}."):
        PopulationDriftDetector(chunk_size=chunk_size, threshold=0.95)

    chunk_size = 0
    with pytest.raises(ValueError, match=f"{error_msg} Got {chunk_size}."):
        PopulationDriftDetector(chunk_size=chunk_size, threshold=0.95)

    chunk_size = "non-valid"
    with pytest.raises(ValueError, match=f"{error_msg} Got {type(chunk_size)}."):
        PopulationDriftDetector(chunk_size=chunk_size, threshold=0.95)


def test_init_exception_when_threshold_not_between_0_and_1():
    """
    Test that an exception is raised when threshold is not between 0 and 1.
    """

    threshold = -0.1
    error_msg = re.escape(f"`threshold` must be between 0 and 1. Got {threshold}.")
    with pytest.raises(ValueError, match=error_msg):
        PopulationDriftDetector(chunk_size='ME', threshold=threshold)

    threshold = 1.1
    error_msg = re.escape(f"`threshold` must be between 0 and 1. Got {threshold}.")
    with pytest.raises(ValueError, match=error_msg):
        PopulationDriftDetector(chunk_size='ME', threshold=threshold)








################################################################################
#                             Population Drift Detector                        #
#                                                                              #
# This work by skforecast team is licensed under the BSD 3-Clause License.     #
################################################################################
# coding=utf-8

from copy import deepcopy
import numpy as np
import pandas as pd
from scipy.stats import chi2_contingency, ecdf
from scipy.spatial.distance import jensenshannon
from sklearn.exceptions import NotFittedError
import warnings

from .. import __version__
from ..exceptions import UnknownLevelWarning
from ..utils import get_style_repr_html


def ks_2samp_from_ecdf(
    ecdf1: object, 
    ecdf2: object, 
    alternative: str = "two-sided"
) -> float:
    """
    Calculate the Kolmogorov-Smirnov distance (statistic) using precomputed
    scipy.stats.ecdf objects. This function replicates the behavior of
    scipy.stats.ks_2samp but uses ECDFs instead of raw data. Only the KS statistic
    is computed, not the p-value.

    Parameters
    ----------
    ecdf1 : scipy.stats.ecdf.ECDF
        ECDF object from sample 1.
    ecdf2 : scipy.stats.ecdf.ECDF
        ECDF object from sample 2.
    alternative : 'two-sided', 'less', 'greater', default 'two-sided'
        Defines the alternative hypothesis.

    Returns
    -------
    distance : float
        KS distance (sup of |F1-F2| for two-sided, etc.).
    
    """

    # Common evaluation grid (all jump points from both ECDFs)
    grid = np.union1d(ecdf1.cdf.quantiles, ecdf2.cdf.quantiles)

    cdf1 = ecdf1.cdf.evaluate(grid)
    cdf2 = ecdf2.cdf.evaluate(grid)

    if alternative == "two-sided":
        distance = np.max(np.abs(cdf1 - cdf2))
    elif alternative == "greater":
        distance = np.max(cdf1 - cdf2)
    elif alternative == "less":
        distance = np.max(cdf2 - cdf1)
    else:
        raise ValueError(
            "Invalid `alternative`. Must be 'two-sided', 'less', or 'greater'."
        )

    return distance


class PopulationDriftDetector:
    """
    A class to detect population drift between reference and new datasets.
    This implementation computes Kolmogorov-Smirnov (KS) test for numeric features,
    Chi-Square test for categorical features, and Jensen-Shannon (JS) distance
    for all features. It calculates empirical distributions of these statistics
    from the reference data and uses quantile thresholds to determine drift in
    new data.
    
    This implementation is inspired by NannyML's DriftDetector. See Notes for
    details.

    For an in-depth explanation of the underlying calculations, see 
    https://skforecast.org/0.18.0/user_guides/drift-detection.html#deep-dive-into-temporal-drift-detection-in-time-series

    Parameters
    ----------
    chunk_size : int, string, pandas DateOffset, default None
        Size of chunks for sequential drift analysis. If int, number of rows per
        chunk. If str (e.g., 'D' for daily, 'W' for weekly), time-based chunks
        assuming a datetime index. If None, analyzes the full dataset as a single
        chunk.
    threshold : float, default 0.95
        The quantile threshold (between 0 and 1) for determining drift based on 
        empirical distributions.
    
    Attributes
    ----------
    chunk_size : int, string, pandas DateOffset
        Size of chunks for sequential drift analysis. If int, number of rows per
        chunk. If str (e.g., 'D' for daily, 'W' for weekly), time-based chunks
        assuming a datetime index. If None, analyzes the full dataset as a single
        chunk.
    threshold : float
        The quantile threshold (between 0 and 1) for determining drift based on 
        empirical distributions.
    is_fitted_ : bool
        Indicates if the detector has been fitted with reference data.
    ref_features_ : list
        List of features in the reference data.
    empirical_dist_ks_ : dict
        Empirical distributions of KS test statistics for each numeric feature in
        reference data.
    empirical_dist_chi2_ : dict
        Empirical distributions of Chi-Square test statistics for each categorical
        feature in reference data.
    empirical_dist_js_ : dict
        Empirical distributions of Jensen-Shannon distance for each feature in
        reference data (numeric and categorical).
    empirical_threshold_ks_ : dict
        Thresholds for KS statistics based on empirical distributions for each
        numeric feature in reference data.
    empirical_threshold_chi2_ : dict
        Thresholds for Chi-Square statistics based on empirical distributions for
        each categorical feature in reference data.
    empirical_threshold_js_ : dict
        Thresholds for Jensen-Shannon distance based on empirical distributions
        for each feature in reference data (numeric and categorical).
    n_chunks_reference_data_ : int
        Number of chunks in the reference data used during fitting to compute
        empirical distributions.
    ref_ecdf_ : dict
        Precomputed ECDFs for numeric features in the reference data.
    ref_bins_edges_ : dict
        Precomputed bin edges for numeric features in the reference data.
    ref_hist_ : dict
        Precomputed histograms for numeric features in the reference data.
    ref_probs_ : dict
        Precomputed normalized value counts (probabilities) for each category of
        categorical features in the reference data.
    ref_ranges_ : dict
        Min and max values for numeric features in the reference data.
    ref_categories_ : dict
        Unique categories for categorical features in the reference data.
    detectors_ : dict
        Dictionary of PopulationDriftDetector instances for each group when
        fitting/predicting on MultiIndex DataFrames.
    series_names_in_ : list
        List of series IDs present during fitting when using MultiIndex DataFrames.
    
    Notes
    -----
    This implementation is inspired by NannyML's DriftDetector [1]_.

    It is a lightweight version adapted for skforecast's needs:
    - It does not store the raw reference data, only the necessary precomputed
    information to calculate the statistics efficiently during prediction.
    - All empirical thresholds are calculated using the specified quantile from
    the empirical distributions obtained from the reference data chunks.
    - It includes checks for out of range values in numeric features and new
    categories in categorical features.
    - It supports multiple time series by fitting separate detectors for each
    series ID when provided with a MultiIndex DataFrame.

    If user requires more advanced features, such as multivariate drift detection
    or data quality checks, consider using https://nannyml.readthedocs.io/en/stable/
    directly.

    References
    ----------
    .. [1] NannyML API Reference.
           https://nannyml.readthedocs.io/en/stable/tutorials/detecting_data_drift/univariate_drift_detection.html
    
    """

    def __init__(
        self, 
        chunk_size=None, 
        threshold=0.95
    ) -> None:
        
        self.ref_features_             = None
        self.is_fitted_                = False
        self.ref_ecdf_                 = {}
        self.ref_bins_edges_           = {}
        self.ref_hist_                 = {}
        self.ref_probs_                = {}
        self.ref_counts_               = {}
        self.empirical_dist_ks_        = {}
        self.empirical_dist_chi2_      = {}
        self.empirical_dist_js_        = {}
        self.empirical_threshold_ks_   = {}
        self.empirical_threshold_chi2_ = {}
        self.empirical_threshold_js_   = {}
        self.ref_ranges_               = {}
        self.ref_categories_           = {}
        self.n_chunks_reference_data_  = None
        self.detectors_                = {}    # NOTE: Only used for multiseries
        self.series_names_in_          = None  # NOTE: Only used for multiseries

        if not (0 < threshold < 1):
            raise ValueError(f"`threshold` must be between 0 and 1. Got {threshold}.")
        
        self.threshold = threshold

        error_msg = (
            "`chunk_size` must be a positive integer, a string compatible with "
            "pandas DateOffset (e.g., 'D', 'W', 'M'), a pandas DateOffset object, or None."
        )
        if not (isinstance(chunk_size, (int, str, pd.DateOffset, type(None)))):
            raise TypeError(f"{error_msg} Got {type(chunk_size)}.")
        
        if isinstance(chunk_size, str):
            try:
                chunk_size = pd.tseries.frequencies.to_offset(chunk_size)
            except ValueError:
                raise ValueError(f"{error_msg} Got {type(chunk_size)}.")
        
        if isinstance(chunk_size, int) and chunk_size <= 0:
            raise ValueError(f"{error_msg} Got {chunk_size}.")
        
        self.chunk_size = chunk_size

    def __repr__(self) -> str:
        """
        Information displayed when a RangeDriftDetector object is printed.
        """
    
        info = (
            f"{'=' * len(type(self).__name__)} \n"
            f"{type(self).__name__} \n"
            f"{'=' * len(type(self).__name__)} \n"
            f"Fitted features = {self.ref_features_} \n"
            f"Is fitted       = {self.is_fitted_}"
        )

        return info
    
    def _repr_html_(self):
        """
        HTML representation of the object.
        The "General Information" section is expanded by default.
        """

        style, unique_id = get_style_repr_html(self.is_fitted_)
        content = f"""
        <div class="container-{unique_id}">
            <p style="font-size: 1.5em; font-weight: bold; margin-block-start: 0.83em; margin-block-end: 0.83em;">{type(self).__name__}</p>
            <details>
                <summary>General Information</summary>
                <ul>
                    <li><strong>Fitted features:</strong> {self.ref_features_}</li>
                    <li><strong>Is fitted:</strong> {self.is_fitted_}</li>
                </ul>
            </details>
            <p>
                <a href="https://skforecast.org/{__version__}/api/drift_detection.html#skforecast.drift_detection._population_drift.PopulationDriftDetector">&#128712 <strong>API Reference</strong></a>
                &nbsp;&nbsp;
                <a href="https://skforecast.org/{__version__}/user_guides/drift-detection.html">&#128462 <strong>User Guide</strong></a>
            </p>
        </div>
        """
        
        return style + content
        
    def _fit(self, X) -> None:
        """
        Fit the drift detector by calculating empirical distributions and thresholds
        from reference data. The empirical distributions are computed by chunking
        the reference data according to the specified `chunk_size` and calculating
        the statistics for each chunk.

        Parameters
        ----------
        X : pandas DataFrame
            Reference data used as the baseline for drift detection.
        
        """

        self.ref_features_             = []
        self.is_fitted_                = False
        self.ref_ecdf_                 = {}
        self.ref_bins_edges_           = {}
        self.ref_hist_                 = {}
        self.ref_probs_                = {}
        self.ref_counts_               = {}
        self.empirical_dist_ks_        = {}
        self.empirical_dist_chi2_      = {}
        self.empirical_dist_js_        = {}
        self.empirical_threshold_ks_   = {}
        self.empirical_threshold_chi2_ = {}
        self.empirical_threshold_js_   = {}
        self.ref_ranges_               = {}
        self.ref_categories_           = {}
        self.n_chunks_reference_data_  = None
        self.detectors_                = {}    # NOTE: Only used for multiseries
        self.series_names_in_          = None  # NOTE: Only used for multiseries

        if self.chunk_size is not None:
            if isinstance(self.chunk_size, pd.offsets.DateOffset) and not isinstance(X.index, pd.DatetimeIndex):
                raise ValueError(
                    "`chunk_size` is a pandas DateOffset but `X` does not have a DatetimeIndex."
                )

        if self.chunk_size is not None:
            if isinstance(self.chunk_size, int):
                chunks_ref = [
                    X.iloc[i : i + self.chunk_size]
                    for i in range(0, len(X), self.chunk_size)
                ]
            elif isinstance(
                self.chunk_size, (str, pd.offsets.DateOffset)
            ) and isinstance(X.index, pd.DatetimeIndex):
                chunks_ref = [group for _, group in X.resample(self.chunk_size)]
        else:
            chunks_ref = [X]

        self.n_chunks_reference_data_ = len(chunks_ref)

        features = X.columns.tolist()
        for feature in features:
            is_numeric = pd.api.types.is_numeric_dtype(X[feature])
            ref = X[feature].dropna()
            if ref.empty:
                warnings.warn(
                    f"Feature '{feature}' contains only NaN values in the reference dataset. "
                    f"Drift detection skipped.",
                    UnknownLevelWarning
                )
                continue

            self.empirical_dist_ks_[feature] = []
            self.empirical_dist_chi2_[feature] = []
            self.empirical_dist_js_[feature] = []
            self.ref_features_.append(feature)

            if is_numeric:
                # Precompute histogram with bins for Jensen-Shannon distance
                # This may not perfectly align with bins used in predict if new data
                # extends the range, but it provides a reasonable approximation
                # for efficiency.
                min_val = ref.min()
                max_val = ref.max()
                bins_edges = np.histogram_bin_edges(ref.astype("float64"), bins='doane')
                ref_hist = np.histogram(ref, bins=bins_edges)[0] / len(ref)
                self.ref_bins_edges_[feature] = bins_edges
                self.ref_hist_[feature] = ref_hist
                self.ref_ranges_[feature] = (min_val, max_val)

                # Precompute ECDF for Kolmogorov-Smirnov test
                self.ref_ecdf_[feature] = ecdf(ref)
            else:
                counts_raw = ref.value_counts()
                counts_norm = counts_raw / counts_raw.sum()
                self.ref_counts_[feature] = counts_raw
                self.ref_probs_[feature] = counts_norm
                self.ref_categories_[feature] = counts_raw.index.tolist()

            for chunk in chunks_ref:
                new = chunk[feature].dropna()
                if new.empty:
                    continue
                ref = ref[~ref.index.isin(new.index)]
                ks_stat = np.nan
                chi2_stat = np.nan
                js_distance = np.nan

                if is_numeric:
                    new_ecdf = ecdf(new)
                    new_hist = np.histogram(new, bins=self.ref_bins_edges_[feature])[0] / len(new)
                    # Handle out-of-bin data: if new data contains values outside the reference range,
                    # they will not be counted in the histogram, leading to a sum < 1. To ensure
                    # the histograms are comparable, we add an extra bin for "out-of-range" data with
                    # the leftover probability mass in the new histogram and a corresponding zero bin
                    # in the reference histogram.
                    leftover = 1 - np.sum(new_hist)
                    if leftover > 0:
                        new_hist = np.append(new_hist, leftover)
                        ref_hist_appended = np.append(self.ref_hist_[feature], 0)
                        js_distance = jensenshannon(ref_hist_appended, new_hist, base=2)
                    else:
                        js_distance = jensenshannon(self.ref_hist_[feature], new_hist, base=2)

                    ks_stat = ks_2samp_from_ecdf(
                        ecdf1=self.ref_ecdf_.get(feature),
                        ecdf2=new_ecdf,
                        alternative="two-sided"
                    )
                else:
                    new_probs = new.value_counts(normalize=True).sort_index()
                    ref_probs = self.ref_probs_.get(feature)
                    # Align categories and fill missing with 0
                    all_cats = ref_probs.index.union(new_probs.index)
                    ref_probs = ref_probs.reindex(all_cats, fill_value=0)
                    new_probs = new_probs.reindex(all_cats, fill_value=0)
                    js_distance = jensenshannon(ref_probs.to_numpy(), new_probs.to_numpy())

                    # Align categories and fill missing with 0
                    new_counts = new.value_counts().reindex(all_cats, fill_value=0).to_numpy()
                    ref_counts = self.ref_counts_.get(feature).reindex(all_cats, fill_value=0).to_numpy()
                    if new_counts.sum() > 0 and ref_counts.sum() > 0:
                        # Create contingency table with rows = [reference, new], columns = categories
                        contingency_table = np.array([ref_counts, new_counts])
                        chi2_stat = chi2_contingency(contingency_table)[0]

                self.empirical_dist_ks_[feature].append(ks_stat)
                self.empirical_dist_chi2_[feature].append(chi2_stat)
                self.empirical_dist_js_[feature].append(js_distance)

            # Calculate empirical thresholds using the the specified quantile
            # Using pandas Series quantile method to handle NaNs properly and warnings
            self.empirical_threshold_ks_[feature] = pd.Series(
                self.empirical_dist_ks_[feature]
            ).quantile(self.threshold)
            self.empirical_threshold_chi2_[feature] = pd.Series(
                self.empirical_dist_chi2_[feature]
            ).quantile(self.threshold)
            self.empirical_threshold_js_[feature] = pd.Series(
                self.empirical_dist_js_[feature]
            ).quantile(self.threshold)

        self.is_fitted_ = True

    def fit(self, X) -> None:
        """
        Fit the drift detector by calculating empirical distributions and thresholds
        from reference data. The empirical distributions are computed by chunking
        the reference data according to the specified `chunk_size` and calculating
        the statistics for each chunk.

        Parameters
        ----------
        X : pandas DataFrame
            Reference data used as the baseline for drift detection.

        """

        if not isinstance(X, pd.DataFrame):
            raise ValueError(
                f"`X` must be a pandas DataFrame. Got {type(X)} instead."
            )

        if isinstance(X.index, pd.MultiIndex):
            X = X.groupby(level=0)

            for idx, group in X:
                group = group.droplevel(0)
                self.detectors_[idx] = PopulationDriftDetector(
                                           chunk_size = self.chunk_size,
                                           threshold  = self.threshold
                                       )
                self.detectors_[idx]._fit(group)
        else:
            self._fit(X)

        self.is_fitted_ = True
        self.series_names_in_ = list(self.detectors_.keys()) if self.detectors_ else None
        self._collect_attributes()

    def _predict(self, X) -> pd.DataFrame:
        """
        Predict drift in new data by comparing the estimated statistics to
        reference thresholds.

        Parameters
        ----------
        X : pandas DataFrame
            New data to compare against the reference.

        Returns
        -------
        results : pandas DataFrame
            DataFrame with the drift detection results for each chunk.

        """
        
        if self.chunk_size is not None:
            if isinstance(self.chunk_size, pd.offsets.DateOffset) and not isinstance(X.index, pd.DatetimeIndex):
                raise ValueError(
                    "`chunk_size` is a pandas DateOffset but `X` does not have a DatetimeIndex."
                )

        if self.chunk_size is not None:
            if isinstance(self.chunk_size, int):
                chunks = [
                    X.iloc[i:i + self.chunk_size]
                    for i in range(0, len(X), self.chunk_size)
                ]
            else:
                chunks = [group for _, group in X.resample(self.chunk_size)]
        else:
            chunks = [X]

        results = []
        features = X.columns.tolist()
        for feature in features:
            if feature not in self.ref_features_:
                warnings.warn(
                    f"Feature '{feature}' was not present during fitting. Drift detection skipped."
                    f"for this feature.",
                    UnknownLevelWarning
                )
                continue

            is_numeric = pd.api.types.is_numeric_dtype(X[feature])
            ref_bin_edges = self.ref_bins_edges_.get(feature, None)
            ref_hist = self.ref_hist_.get(feature, None)
            ref_probs = self.ref_probs_.get(feature, None)
            ref_counts = self.ref_counts_.get(feature, None)
            ref_ecdf = self.ref_ecdf_.get(feature, None)
            threshold_ks = self.empirical_threshold_ks_.get(feature, np.nan)
            threshold_chi2 = self.empirical_threshold_chi2_.get(feature, np.nan)
            threshold_js = self.empirical_threshold_js_.get(feature, np.nan)
            ref_range = self.ref_ranges_.get(feature, (np.nan, np.nan))

            for chunk_idx, chunk in enumerate(chunks):

                chunk_label = chunk_idx if self.chunk_size else "full"
                new = chunk[feature].dropna()
                ks_stat = np.nan
                chi2_stat = np.nan
                js_distance = np.nan
                is_out_of_range = np.nan

                if not new.empty:

                    if is_numeric:
                        new_ecdf = ecdf(new)
                        # Compute histogram for new data using reference bin edges and normalize
                        new_hist = np.histogram(new, bins=ref_bin_edges)[0] / len(new)
                        # Handle out-of-bin data: if new data contains values outside the reference range,
                        # they will not be counted in the histogram, leading to a sum < 1. To ensure
                        # the histograms are comparable, we add an extra bin for "out-of-range" data with
                        # the leftover probability mass in the new histogram and a corresponding zero bin
                        # in the reference histogram.
                        leftover = 1 - np.sum(new_hist)
                        if leftover > 0:
                            new_hist = np.append(new_hist, leftover)
                            ref_hist_appended = np.append(ref_hist, 0)
                            js_distance = jensenshannon(ref_hist_appended, new_hist, base=2)
                        else:
                            js_distance = jensenshannon(ref_hist, new_hist, base=2)

                        ks_stat = ks_2samp_from_ecdf(
                            ecdf1=ref_ecdf,
                            ecdf2=new_ecdf,
                            alternative="two-sided"
                        )
                        is_out_of_range = (
                            np.min(new) < ref_range[0] or
                            np.max(new) > ref_range[1]
                        )
                    else:
                        ref_categories = self.ref_categories_[feature]
                        ref_probs_ = ref_probs.reindex(ref_categories, fill_value=0).to_numpy()
                        # Map new data to reference categories
                        new_counts_dict = new.value_counts().to_dict()
                        new_counts_on_ref = [new_counts_dict.get(cat, 0) for cat in ref_categories]
                        new_probs = (
                            np.array(new_counts_on_ref) / len(new) if len(new) > 0
                            else np.zeros(len(ref_categories))
                        )
                        # Compute leftover (probability of new categories not in reference): if new data
                        # contains categories not seen in reference, they will not be counted in the
                        # histogram, leading to a sum < 1. To ensure the histograms are comparable,
                        # we add an extra bin for "new categories" with the leftover probability mass
                        # in the new histogram and a corresponding zero bin in the reference histogram.
                        leftover = 1 - np.sum(new_probs)
                        if leftover > 0:
                            new_probs = np.append(new_probs, leftover)
                            ref_probs_appended = np.append(ref_probs_, 0)
                            js_distance = jensenshannon(ref_probs_appended, new_probs, base=2)
                        else:
                            js_distance = jensenshannon(ref_probs_, new_probs, base=2)

                        all_cats = set(self.ref_categories_[feature]).union(set(new_counts_dict.keys()))
                        new_counts = new.value_counts().reindex(all_cats, fill_value=0).to_numpy()
                        ref_counts_aligned = ref_counts.reindex(all_cats, fill_value=0).to_numpy()
                        if new_counts.sum() > 0 and ref_counts_aligned.sum() > 0:
                            # Create contingency table: rows = [reference, new], columns = categories
                            contingency_table = np.array([ref_counts_aligned, new_counts])
                            chi2_stat = chi2_contingency(contingency_table)[0]

                results.append({
                    "chunk": chunk_label,
                    "chunk_start": chunk.index.min(),
                    "chunk_end": chunk.index.max(),
                    "feature": feature,
                    "ks_statistic": ks_stat,
                    "threshold_ks": threshold_ks,
                    "chi2_statistic": chi2_stat,
                    "threshold_chi2": threshold_chi2,
                    "jensen_shannon": js_distance,
                    "threshold_js": threshold_js,
                    "reference_range": ref_range,
                    "is_out_of_range": is_out_of_range,
                })

        results_df = pd.DataFrame(results)
        results_df['drift_ks_statistic'] = results_df['ks_statistic'] > results_df['threshold_ks']
        results_df['drift_chi2_statistic'] = results_df['chi2_statistic'] > results_df['threshold_chi2']
        results_df['drift_js'] = results_df['jensen_shannon'] > results_df['threshold_js']
        results_df['drift_detected'] = (
            results_df['drift_ks_statistic']
            | results_df['drift_chi2_statistic']
            | results_df['drift_js']
            | results_df['is_out_of_range']
        )
        
        return results_df

    def predict(self, X) -> tuple[pd.DataFrame, pd.DataFrame]:
        """
        Predict drift in new data by comparing the estimated statistics to
        reference thresholds. Two dataframes are returned, the first one with
        detailed information of each chunk, the second only the total number
        of chunks where drift have been detected.

        Parameters
        ----------
        X : pandas DataFrame
            New data to compare against the reference.

        Returns
        -------
        results : pandas DataFrame
            DataFrame with the drift detection results for each chunk.
        summary : pandas DataFrame
            Summary DataFrame with the total number and percentage of chunks
            with detected drift per feature (or per series_id and feature if
            MultiIndex).
        
        """
        
        if not self.is_fitted_:
            raise NotFittedError(
                "This PopulationDriftDetector instance is not fitted yet. "
                "Call 'fit' with appropriate arguments before using this estimator."
            )

        if not isinstance(X, pd.DataFrame):
            raise ValueError(f"`X` must be a pandas DataFrame. Got {type(X)} instead.")

        if isinstance(X.index, pd.MultiIndex):
            results = []
            for idx, group in X.groupby(level=0):
                group = group.droplevel(0)
                if idx not in self.detectors_:
                    warnings.warn(
                        f"Series '{idx}' was not present during fitting. Drift detection skipped.",
                        UnknownLevelWarning
                    )
                    continue

                detector = self.detectors_[idx]
                result = detector._predict(group)
                result.insert(0, 'series_id', idx)
                results.append(result)

            results = pd.concat(results, ignore_index=True)
        else:
            results = self._predict(X)

        if results.columns[0] == 'series_id':
            summary = (
                results.groupby(['series_id', 'feature'])['drift_detected']
                .agg(['sum', 'mean'])
                .reset_index()
                .rename(columns={'sum': 'n_chunks_with_drift', 'mean': 'pct_chunks_with_drift'})
            )
        else:
            summary = (
                results.groupby(['feature'])['drift_detected']
                .agg(['sum', 'mean'])
                .reset_index()
                .rename(columns={'sum': 'n_chunks_with_drift', 'mean': 'pct_chunks_with_drift'})
            )
        summary['pct_chunks_with_drift'] = summary['pct_chunks_with_drift'] * 100
        
        return results, summary

    def _collect_attributes(self) -> None:
        """
        Collect attributes for representation and inspection and update the instance
        dictionary with the collected values. For multi-series (when detectors_ is
        populated), attributes are aggregated into nested dictionaries keyed by
        detector names. For single-series, attributes remain unchanged.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """

        attr_names = [
            k 
            for k in self.__dict__.keys() 
            if k not in ['is_fitted_', 'detectors_', 'series_names_in_']
        ]
        
        if self.detectors_:
            for attr_name in attr_names:
                collected = {}
                for detector_key, detector in self.detectors_.items():
                    collected[detector_key] = getattr(detector, attr_name, None)
                self.__dict__[attr_name] = deepcopy(collected)



# Unit test check_backtesting_input
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.linear_model import Ridge, LogisticRegression
from sklearn.exceptions import NotFittedError
from skforecast.stats import Sarimax
from skforecast.recursive import ForecasterRecursive
from skforecast.recursive import ForecasterRecursiveClassifier
from skforecast.direct import ForecasterDirect
from skforecast.recursive import ForecasterStats
from skforecast.recursive import ForecasterEquivalentDate
from skforecast.recursive import ForecasterRecursiveMultiSeries
from skforecast.direct import ForecasterDirectMultiVariate
from skforecast.model_selection._split import TimeSeriesFold
from skforecast.model_selection._utils import check_backtesting_input

# Fixtures
from skforecast.model_selection.tests.fixtures_model_selection import y
from skforecast.model_selection.tests.fixtures_model_selection_multiseries import (
    series_wide_range,
    series_dict_range,
    series_dict_dt
)


def test_check_backtesting_input_TypeError_when_cv_not_TimeSeriesFold():
    """
    Test TypeError is raised in check_backtesting_input if `cv` is not a
    TimeSeriesFold object.
    """
    forecaster = ForecasterRecursive(estimator=Ridge(), lags=2)
    y = pd.Series(np.arange(50))
    y.index = pd.date_range(start='2000-01-01', periods=len(y), freq='D')
    
    class BadCv():
        pass

    err_msg = re.escape("`cv` must be a 'TimeSeriesFold' object. Got 'BadCv'.")
    with pytest.raises(TypeError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = BadCv(),
            metric                  = 'mean_absolute_error',
            y                       = y,
            series                  = None,
            interval                = None,
            alpha                   = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


@pytest.mark.parametrize("forecaster", 
                         [ForecasterRecursive(estimator=Ridge(), lags=2),
                          ForecasterDirect(estimator=Ridge(), lags=2, steps=3),
                          ForecasterStats(estimator=Sarimax(order=(1, 1, 1)))], 
                         ids = lambda fr: f'forecaster: {type(fr).__name__}')
def test_check_backtesting_input_TypeError_when_y_is_not_pandas_Series(forecaster):
    """
    Test TypeError is raised in check_backtesting_input if `y` is not a 
    pandas Series in forecasters uni-series.
    """
    bad_y = np.arange(50)
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(bad_y) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True,
             verbose               = False
         )

    err_msg = re.escape("`y` must be a pandas Series.")
    with pytest.raises(TypeError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            y                       = bad_y,
            series                  = None,
            interval                = None,
            alpha                   = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


@pytest.mark.parametrize("forecaster", 
                         [ForecasterDirectMultiVariate(estimator=Ridge(), lags=2, 
                                                        steps=3, level='l1')], 
                         ids = lambda fr: f'forecaster: {type(fr).__name__}')
def test_check_backtesting_input_TypeError_when_series_not_pandas_DataFrame(forecaster):
    """
    Test TypeError is raised in check_backtesting_input if `series` is not a 
    pandas DataFrame in forecasters multiseries.
    """
    bad_series = pd.Series(np.arange(50))
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(bad_series) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True,
             verbose               = False
         )

    err_msg = re.escape("`series` must be a pandas DataFrame.")
    with pytest.raises(TypeError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            series                  = bad_series,
            interval                = None,
            alpha                   = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


def test_check_backtesting_input_TypeError_when_not_valid_exog_type():
    """
    Test TypeError is raised in check_backtesting_input if `exog` is not a
    pandas Series, DataFrame or None.
    """
    y = pd.Series(np.arange(50))
    y.index = pd.date_range(start='2000-01-01', periods=len(y), freq='D')

    forecaster = ForecasterRecursive(estimator=Ridge(), lags=2)

    bad_exog = np.arange(50)
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y[:-12]),
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True,
             verbose               = False
         )

    err_msg = re.escape(
        f"`exog` must be a pandas Series, DataFrame or None. Got {type(bad_exog)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            y                       = y,
            exog                    = bad_exog,
            interval                = None,
            alpha                   = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


@pytest.mark.parametrize("differentiation", 
    [{'l1': 1, 'l2': 2, '_unknown_level': 1}, {'l1': 2, 'l2': None, '_unknown_level': 1}], 
     ids = lambda diff: f'differentiation: {diff}')
def test_check_backtesting_input_ValueError_when_ForecasterRecursiveMultiSeries_diff_dict_not_cv_diff(differentiation):
    """
    Test ValueError is raised in check_backtesting_input if `differentiation`
    of the ForecasterRecursiveMultiSeries as dict is different from 
    `differentiation` of the cv.
    """
    forecaster = ForecasterRecursiveMultiSeries(
        estimator=Ridge(), lags=2, differentiation=differentiation
    )

    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(series_dict_range['l1']) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True,
             differentiation       = 1,
             verbose               = False
         )
    
    err_msg = re.escape(
        "When using a dict as `differentiation` in ForecasterRecursiveMultiSeries, "
        "the `differentiation` included in the cv (1) must be "
        "the same as the maximum `differentiation` included in the forecaster "
        "(2). Set the same value "
        "for both using the `differentiation` argument."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            series                  = series_dict_range,
            interval                = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


@pytest.mark.parametrize("forecaster", 
    [ForecasterRecursive(estimator=Ridge(), lags=2, differentiation=2),
     ForecasterRecursiveMultiSeries(estimator=Ridge(), lags=2, differentiation=2)], 
     ids = lambda fr: f'forecaster: {type(fr).__name__}')
def test_check_backtesting_input_ValueError_when_forecaster_diff_not_cv_diff(forecaster):
    """
    Test ValueError is raised in check_backtesting_input if `differentiation`
    of the forecaster is different from `differentiation` of the cv.
    """
    if type(forecaster).__name__ == 'ForecasterRecursive':
        data_length = len(y)
    else:
        data_length = len(series_dict_range['l1'])
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = data_length - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True,
             differentiation       = 1,
             verbose               = False
         )
    
    err_msg = re.escape(
        "The differentiation included in the forecaster "
        "(2) differs from the differentiation "
        "included in the cv (1). Set the same value "
        "for both using the `differentiation` argument."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            y                       = y,
            series                  = series_dict_range,
            interval                = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


def test_check_backtesting_input_TypeError_when_metric_not_correct_type():
    """
    Test TypeError is raised in check_backtesting_input if `metric` is not string, 
    a callable function, or a list containing multiple strings and/or callables.
    """
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 2
                 )
    
    cv = TimeSeriesFold(
             steps                 = 5,
             initial_train_size    = len(y[:-12]),
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    metric = 5
    
    err_msg = re.escape(
        f"`metric` must be a string, a callable function, or a list containing "
        f"multiple strings and/or callables. Got {type(metric)}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = metric,
            y                       = y,
            interval                = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


def test_check_backtesting_input_ValueError_when_initial_train_size_is_None_ForecasterEquivalentDate():
    """
    Test ValueError is raised in check_backtesting_input when initial_train_size
    is None with a ForecasterEquivalentDate with a offset of type pd.DateOffset.
    """

    data_length = len(y)
    data_name = 'y'

    forecaster = ForecasterEquivalentDate(
                     offset    = pd.DateOffset(days=10),
                     n_offsets = 2 
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = None,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    err_msg = re.escape(
        f"`initial_train_size` must be an integer greater than "
        f"the `window_size` of the forecaster ({forecaster.window_size}) "
        f"and smaller than the length of `{data_name}` ({data_length}) or "
        f"a date within this range of the index."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            y                       = y,
            interval                = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


@pytest.mark.parametrize("initial_train_size", 
                         ['greater', 'smaller', 'date'], 
                         ids = lambda initial: f'initial_train_size: {initial}')
@pytest.mark.parametrize("forecaster", 
                         [ForecasterRecursive(estimator=Ridge(), lags=3),
                          ForecasterRecursiveMultiSeries(estimator=Ridge(), lags=3)], 
                         ids = lambda fr: f'{type(fr).__name__}')
def test_check_backtesting_input_ValueError_when_initial_train_size_not_correct_value(initial_train_size, forecaster):
    """
    Test ValueError is raised in check_backtesting_input when 
    initial_train_size >= length `y` or `series` or initial_train_size < window_size.
    """
    y_datetime = y.copy()
    y_datetime.index = pd.date_range(start='2020-01-01', periods=len(y), freq='D')

    if type(forecaster).__name__ == 'ForecasterRecursive':
        data_length = len(y_datetime)
        data_name = 'y'
    else:
        data_length = len(series_dict_dt['l1'])
        data_name = 'series'

    if initial_train_size == 'greater':
        initial_train_size = data_length
    elif initial_train_size == 'smaller':
        initial_train_size = forecaster.window_size - 1
    else:
        initial_train_size = '2020-01-02'  # Smaller than window_size
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = initial_train_size,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    err_msg = re.escape(
        f"If `initial_train_size` is an integer, it must be greater than "
        f"the `window_size` of the forecaster ({forecaster.window_size}) "
        f"and smaller than the length of `{data_name}` ({data_length}). If "
        f"it is a date, it must be within this range of the index."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            y                       = y_datetime,
            series                  = series_dict_dt,
            interval                = None,
            alpha                   = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


@pytest.mark.parametrize("initial_train_size", 
                         ['int', 'date'], 
                         ids = lambda initial: f'initial_train_size: {initial}')
@pytest.mark.parametrize("forecaster", 
                         [ForecasterRecursive(estimator=Ridge(), lags=2),
                          ForecasterRecursiveMultiSeries(estimator=Ridge(), lags=2)], 
                         ids = lambda fr: f'{type(fr).__name__}')
def test_check_backtesting_input_ValueError_when_not_enough_data_to_create_a_fold_allow_incomplete_fold(initial_train_size, forecaster):
    """
    Test ValueError is raised in check_backtesting_input when there is not enough 
    data to evaluate even single fold because `allow_incomplete_fold` = `True`.
    """
    y_datetime = y.copy()
    y_datetime.index = pd.date_range(start='2000-01-01', periods=len(y), freq='D')

    if type(forecaster).__name__ == 'ForecasterRecursive':
        data_length = len(y_datetime)
        data_name = 'y'
    else:
        data_length = len(series_dict_dt['l1'])
        data_name = 'series'

    if initial_train_size == 'int':
        initial_train_size = data_length - 1
    else:
        initial_train_size = '2000-02-19'
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = data_length - 1,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 2,
             allow_incomplete_fold = True
         )
    
    err_msg = re.escape(
        f"`{data_name}` must have more than `initial_train_size + gap` "
        f"observations to create at least one fold.\n"
        f"    Time series length: {data_length}\n"
        f"    Required > {cv.initial_train_size + cv.gap}\n"
        f"    initial_train_size: {cv.initial_train_size}\n"
        f"    gap: {cv.gap}\n"
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            y                       = y,
            series                  = series_dict_range,
            interval                = None,
            alpha                   = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


@pytest.mark.parametrize("initial_train_size", 
                         ['int', 'date'], 
                         ids = lambda initial: f'initial_train_size: {initial}')
@pytest.mark.parametrize("forecaster", 
                         [ForecasterRecursive(estimator=Ridge(), lags=2),
                          ForecasterRecursiveMultiSeries(estimator=Ridge(), lags=2)], 
                         ids = lambda fr: f'{type(fr).__name__}')
def test_check_backtesting_input_ValueError_when_not_enough_data_to_create_a_fold_allow_incomplete_fold_False(initial_train_size, forecaster):
    """
    Test ValueError is raised in check_backtesting_input when there is not enough 
    data to evaluate even single fold because `allow_incomplete_fold` = `False`.
    """
    y_datetime = y.copy()
    y_datetime.index = pd.date_range(start='2000-01-01', periods=len(y), freq='D')

    if type(forecaster).__name__ == 'ForecasterRecursive':
        data_length = len(y_datetime)
        data_name = 'y'
    else:
        data_length = len(series_dict_dt['l1'])
        data_name = 'series'

    if initial_train_size == 'int':
        initial_train_size = data_length - 1
    else:
        initial_train_size = '2000-02-19'
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = data_length - 1,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 2,
             allow_incomplete_fold = False
         )
    
    err_msg = re.escape(
        f"`{data_name}` must have at least `initial_train_size + gap + steps` "
        f"observations to create a minimum of one complete fold "
        f"(allow_incomplete_fold=False).\n"
        f"    Time series length: {data_length}\n"
        f"    Required >= {cv.initial_train_size + cv.gap + cv.steps}\n"
        f"    initial_train_size: {cv.initial_train_size}\n"
        f"    gap: {cv.gap}\n"
        f"    steps: {cv.steps}\n"
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            y                       = y,
            series                  = series_dict_range,
            interval                = None,
            alpha                   = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


@pytest.mark.parametrize("forecaster", 
                         [ForecasterStats(estimator=Sarimax(order=(1, 1, 1))),
                          ForecasterEquivalentDate(offset=1, n_offsets=1)], 
                         ids = lambda fr: f'{type(fr).__name__}')
def test_check_backtesting_input_ValueError_Sarimax_Equivalent_when_initial_train_size_is_None(forecaster):
    """
    Test ValueError is raised in check_backtesting_input when initial_train_size 
    is None with a ForecasterStats or ForecasterEquivalentDate.
    """
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = None,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    err_msg = re.escape(
        f"`initial_train_size` must be an integer smaller than the "
        f"length of `y` ({len(y)})."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            y                       = y,
            interval                = None,
            alpha                   = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


def test_check_backtesting_input_NotFittedError_when_initial_train_size_None_and_forecaster_not_fitted():
    """
    Test NotFittedError is raised in check_backtesting_input when 
    initial_train_size is None and forecaster is not fitted.
    """
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 2
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = None,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    err_msg = re.escape(
        "`forecaster` must be already trained if no `initial_train_size` "
        "is provided."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            y                       = y,
            interval                = None,
            alpha                   = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


def test_check_backtesting_input_ValueError_when_initial_train_size_None_and_refit_True():
    """
    Test ValueError is raised in check_backtesting_input when initial_train_size 
    is None and refit is True.
    """
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 2
                 )
    forecaster.is_fitted = True
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = None,
             refit                 = True,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    err_msg = re.escape(
        "`refit` is only allowed when `initial_train_size` is not `None`."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            y                       = y,
            interval                = None,
            alpha                   = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


def test_check_backtesting_input_ValueError_when_skip_folds_in_ForecasterStats():
    """
    Test ValueError is raised in check_backtesting_input if `skip_folds` is
    used in ForecasterStats.
    """
    forecaster = ForecasterStats(
                     estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True,
             skip_folds            = 2,
         )
    
    err_msg = re.escape(
        "`skip_folds` is not allowed for ForecasterStats. Set it to `None`."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            y                       = y,
            interval                = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False
        )


@pytest.mark.parametrize("boolean_argument", 
                         ['add_aggregated_metric', 'use_in_sample_residuals', 
                          'use_binned_residuals', 'return_predictors', 'show_progress', 
                          'suppress_warnings', 'suppress_warnings_fit'], 
                         ids = lambda argument: f'{argument}')
def test_check_backtesting_input_TypeError_when_boolean_arguments_not_bool(boolean_argument):
    """
    Test TypeError is raised in check_backtesting_input when boolean arguments 
    are not boolean.
    """
    forecaster = ForecasterRecursive(
                    estimator = Ridge(random_state=123),
                    lags      = 2
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    boolean_arguments = {
        'add_aggregated_metric': False,
        'use_in_sample_residuals': False,
        'use_binned_residuals': False,
        'return_predictors': False,
        'show_progress': False,
        'suppress_warnings': False,
        'suppress_warnings_fit': False
    }
    boolean_arguments[boolean_argument] = 'not_bool'
    
    err_msg = re.escape(f"`{boolean_argument}` must be a boolean: `True`, `False`.")
    with pytest.raises(TypeError, match = err_msg):
        check_backtesting_input(
            forecaster   = forecaster,
            cv           = cv,
            metric       = 'mean_absolute_error',
            y            = y,
            interval     = None,
            alpha        = None,
            n_boot       = 500,
            random_state = 123,
            **boolean_arguments
        )


@pytest.mark.parametrize("int_argument, value",
                         [('n_boot', 2.2), 
                          ('n_boot', -2),
                          ('random_state', 'not_int'),  
                          ('random_state', -3)], 
                         ids = lambda argument: f'{argument}')
def test_check_backtesting_input_TypeError_when_integer_args_not_int_or_greater_than_0(int_argument, value):
    """
    Test TypeError is raised in check_backtesting_input when integer arguments 
    are not int or are greater than 0.
    """
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 2
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    integer_arguments = {'n_boot': 500, 'random_state': 123}
    integer_arguments[int_argument] = value
    
    err_msg = re.escape(f"`{int_argument}` must be an integer greater than 0. Got {value}.")
    with pytest.raises(TypeError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            y                       = y,
            interval                = None,
            alpha                   = None,
            use_in_sample_residuals = True,
            show_progress           = False,
            suppress_warnings       = False,
            **integer_arguments
        )


@pytest.mark.parametrize("n_jobs", 
                         [1.0, 'not_int_auto'], 
                         ids = lambda value: f'n_jobs: {value}')
def test_check_backtesting_input_TypeError_when_n_jobs_not_int_or_auto(n_jobs):
    """
    Test TypeError is raised in check_backtesting_input when n_jobs  
    is not an integer or 'auto'.
    """
    forecaster = ForecasterRecursive(
                     estimator = Ridge(random_state=123),
                     lags      = 2
                 )
    
    cv = TimeSeriesFold(
             steps                 = 3,
             initial_train_size    = len(y) - 12,
             refit                 = False,
             fixed_train_size      = False,
             gap                   = 0,
             allow_incomplete_fold = True
         )
    
    err_msg = re.escape(f"`n_jobs` must be an integer or `'auto'`. Got {n_jobs}.")
    with pytest.raises(TypeError, match = err_msg):
        check_backtesting_input(
            forecaster              = forecaster,
            cv                      = cv,
            metric                  = 'mean_absolute_error',
            y                       = y,
            interval                = None,
            n_boot                  = 500,
            random_state            = 123,
            use_in_sample_residuals = True,
            n_jobs                  = n_jobs,
            show_progress           = False,
            suppress_warnings       = False
        )


def test_check_backtesting_input_raises_when_interval_not_None_and_interval_method():
    """
    Test raises errors in check_backtesting_input when interval is not None
    and the forecaster uses bootstrapping or conformal.
    """
    forecaster = ForecasterRecursive(estimator=Ridge(), lags=2)
    cv = TimeSeriesFold(steps=3, initial_train_size=len(y) - 12)
    
    kwargs = {
        'forecaster': forecaster,
        'cv': cv,
        'metric': 'mean_absolute_error',
        'y': y,
        'interval': [10, 90],
        'interval_method': 'bootstrapping',
        'n_boot': 500,
        'use_in_sample_residuals': True,
        'use_binned_residuals': False,
        'random_state': 123,
        'show_progress': False,
        'suppress_warnings': False
    }

    kwargs['interval'] = {'10': 10, '90': 90}
    kwargs['interval_method'] = 'conformal'
    err_msg = re.escape(
        f"When `interval_method` is 'conformal', `interval` must "
        f"be a float or a list/tuple defining a symmetric interval. "
        f"Got {type(kwargs['interval'])}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_backtesting_input(**kwargs)

    kwargs['interval'] = {'10': 10, '90': 90}
    kwargs['interval_method'] = 'bootstrapping'
    err_msg = re.escape(
        f"When `interval_method` is 'bootstrapping', `interval` "
        f"must be a float, a list or tuple of floats, a "
        f"scipy.stats distribution object (with methods `_pdf` and "
        f"`fit`) or the string 'bootstrapping'. Got {type(kwargs['interval'])}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_backtesting_input(**kwargs)

    class CustomObject:  # pragma: no cover
        pass
    
    kwargs['interval'] = CustomObject()
    err_msg = re.escape(
        f"When `interval_method` is 'bootstrapping', `interval` "
        f"must be a float, a list or tuple of floats, a "
        f"scipy.stats distribution object (with methods `_pdf` and "
        f"`fit`) or the string 'bootstrapping'. Got {type(kwargs['interval'])}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_backtesting_input(**kwargs)

    kwargs['interval'] = ['10', '90']
    err_msg = re.escape(
        f"`interval` must be a list or tuple of floats. "
        f"Got {type('10')} in {kwargs['interval']}."
    )
    with pytest.raises(TypeError, match = err_msg):
        check_backtesting_input(**kwargs)

    kwargs['interval'] = [0, 100, 101]
    err_msg = re.escape(
        "When `interval` is a list or tuple, all values must be "
        "between 0 and 100 inclusive."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(**kwargs)

    kwargs['interval'] = 'not_bootstrapping'
    err_msg = re.escape(
        f"When `interval` is a string, it must be 'bootstrapping'."
        f"Got {kwargs['interval']}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(**kwargs)

    kwargs['interval_method'] = 'not_bootstrapping_or_conformal'
    err_msg = re.escape(
        f"`interval_method` must be 'bootstrapping' or 'conformal'. "
        f"Got {kwargs['interval_method']}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(**kwargs)


def test_check_backtesting_input_ValueError_when_interval_and_ForecasterRecursiveClassifier():
    """
    Test ValueError is raised in check_backtesting_input when `interval` is not None
    and the forecaster is a ForecasterRecursiveClassifier.
    """
    forecaster = ForecasterRecursiveClassifier(
        estimator=LogisticRegression(), lags=2
    )
    
    cv = TimeSeriesFold(
             steps              = 3,
             initial_train_size = len(y) - 12,
         )
    
    err_msg = re.escape(
        "`interval` is not supported for ForecasterRecursiveClassifier. Class "
        "probabilities are returned by default during backtesting, "
        "set `interval=None`."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(
            forecaster        = forecaster,
            cv                = cv,
            metric            = 'mean_absolute_error',
            y                 = y,
            interval         = [10, 90],
            show_progress     = False,
            suppress_warnings = False
        )


def test_check_backtesting_input_ValueError_when_return_predictors_and_forecaster_not_return_predictors():
    """
    Test ValueError is raised in check_backtesting_input when `return_predictors` is True
    and the forecaster is not an allowed type.
    """
    forecaster = ForecasterStats(
        estimator = Sarimax(order=(3, 2, 0), maxiter=1000, method='cg', disp=False)
    )
    
    cv = TimeSeriesFold(
             steps              = 3,
             initial_train_size = len(y) - 12,
         )
    
    forecaster_name = type(forecaster).__name__
    forecasters_return_predictors = [
        "ForecasterRecursive",
        "ForecasterDirect",
        "ForecasterRecursiveMultiSeries",
        "ForecasterDirectMultiVariate",
        "ForecasterRecursiveClassifier"
    ]
    
    err_msg = re.escape(
        f"`return_predictors` is only allowed for forecasters of type "
        f"{forecasters_return_predictors}. Got {forecaster_name}."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(
            forecaster        = forecaster,
            cv                = cv,
            metric            = 'mean_absolute_error',
            y                 = y,
            return_predictors = True,
            show_progress     = False,
            suppress_warnings = False
        )


@pytest.mark.parametrize("forecaster", 
                         [ForecasterDirect(estimator=Ridge(), lags=5, steps=5),
                          ForecasterDirectMultiVariate(estimator=Ridge(), level='l1', lags=5, steps=5)], 
                         ids = lambda fr: f'{type(fr).__name__}')
def test_check_backtesting_input_ValueError_when_Direct_forecaster_not_enough_steps(forecaster):
    """
    Test ValueError is raised in check_backtesting_input when there is not enough 
    steps to predict steps + gap in a Direct forecaster.
    """
    
    cv = TimeSeriesFold(
             steps                 = 1,
             initial_train_size    = len(y[:-12]),
             window_size           = None,
             differentiation       = None,
             refit                 = False,
             fixed_train_size      = True,
             gap                   = 5,
             skip_folds            = None,
             allow_incomplete_fold = True,
             return_all_indexes    = False,
         )
    
    err_msg = re.escape(
        f"When using a {type(forecaster).__name__}, the combination of steps "
        f"+ gap ({cv.steps + cv.gap}) cannot be greater than the `steps` parameter "
        f"declared when the forecaster is initialized ({forecaster.max_step})."
    )
    with pytest.raises(ValueError, match = err_msg):
        check_backtesting_input(
            forecaster        = forecaster,
            cv                = cv,
            metric            = 'mean_absolute_error',
            y                 = y,
            series            = series_wide_range,
            show_progress     = False,
            suppress_warnings = False
        )



# Unit test predict ForecasterRecursiveClassifier
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.preprocessing import FunctionTransformer
from sklearn.compose import make_column_transformer
from sklearn.compose import make_column_selector
from sklearn.pipeline import make_pipeline
from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import HistGradientBoostingClassifier
from lightgbm import LGBMClassifier

from skforecast.preprocessing import RollingFeaturesClassification
from skforecast.recursive import ForecasterRecursiveClassifier

# Fixtures
from .fixtures_forecaster_recursive_classifier import y, y_dt
from .fixtures_forecaster_recursive_classifier import exog, exog_dt, exog_predict, exog_dt_predict


def test_predict_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=3)

    err_msg = re.escape(
        "This Forecaster instance is not fitted yet. Call `fit` with "
        "appropriate arguments before using predict."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster.predict(steps=5)


def test_predict_output_when_estimator_is_LogisticRegression():
    """
    Test predict output when using LogisticRegression as estimator.
    """
    y_dummy = pd.Series(
        np.array(['a', 'b', 'c', 'a', 'b', 'c', 'a', 'b', 'c', 'a', 'b', 'c', 'a', 'b', 'c']), 
        index=pd.date_range("2020-01-01", periods=15),
        name='y'
    )

    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=3)
    forecaster.fit(y=y_dummy)
    predictions = forecaster.predict(steps=5)

    expected = pd.Series(
                   data = np.array(['a', 'b', 'c', 'a', 'b']),
                   index = pd.date_range("2020-01-16", periods=5),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)

        
def test_predict_output_when_with_exog():
    """
    Test predict output when using LogisticRegression as estimator.
    """
    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=3)
    forecaster.fit(y=y, exog=exog)
    
    predictions = forecaster.predict(steps=5, exog=exog_predict)

    expected = pd.Series(
                   data = np.array([2, 2, 2, 2, 2]),
                   index = pd.RangeIndex(start=50, stop=55, step=1),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


def test_predict_output_with_transform_exog():
    """
    Test predict output when using LogisticRegression as estimator and 
    transformer_exog.
    """

    df_exog = pd.DataFrame({
               'col_1': exog.to_numpy(),
               'col_2': ['a', 'a', 'b', 'a', 'b'] * 10}
           )
    df_exog_predict = df_exog.iloc[:10, :].copy()
    df_exog_predict.index = pd.RangeIndex(start=50, stop=60)

    transformer_exog = ColumnTransformer(
                            [('scale', StandardScaler(), ['col_1']),
                             ('onehot', OneHotEncoder(), ['col_2'])],
                            remainder = 'passthrough',
                            verbose_feature_names_out = False
                       )
    forecaster = ForecasterRecursiveClassifier(
                     estimator        = LogisticRegression(),
                     lags             = 5,
                     transformer_exog = transformer_exog,
                 )
    forecaster.fit(y=y, exog=df_exog)
    predictions = forecaster.predict(steps=5, exog=df_exog_predict)

    expected = pd.Series(
                   data = np.array([3, 1, 2, 3, 2]),
                   index = pd.RangeIndex(start=50, stop=55, step=1),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


def test_predict_output_when_categorical_features_native_implementation_HistGradientBoostingClassifier():
    """
    Test predict output when using HistGradientBoostingClassifier and categorical variables.
    """
    df_exog = pd.DataFrame({'exog_1': exog.to_numpy(),
                            'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)})
    
    df_exog_predict = df_exog.iloc[:10, :].copy()
    df_exog_predict.index = pd.RangeIndex(start=50, stop=60)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterRecursiveClassifier(
                     estimator        = HistGradientBoostingClassifier(
                                            categorical_features = categorical_features,
                                            random_state         = 123
                                        ),
                     lags             = 5,
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y, exog=df_exog)
    predictions = forecaster.predict(steps=10, exog=df_exog_predict)

    expected = pd.Series(
                   data = np.array([2, 2, 2, 2, 2, 2, 2, 2, 2, 1]),
                   index = pd.RangeIndex(start=50, stop=60, step=1),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


def test_predict_output_when_categorical_features_native_implementation_LGBMClassifier():
    """
    Test predict output when using LGBMClassifier and categorical variables.
    """
    df_exog = pd.DataFrame({'exog_1': exog.to_numpy(),
                            'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)})
    
    df_exog_predict = df_exog.iloc[:10, :].copy()
    df_exog_predict.index = pd.RangeIndex(start=50, stop=60)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterRecursiveClassifier(
                     estimator        = LGBMClassifier(verbose=-1, random_state=123),
                     lags             = 5,
                     transformer_exog = transformer_exog,
                     fit_kwargs       = {'categorical_feature': categorical_features}
                 )
    forecaster.fit(y=y, exog=df_exog)
    predictions = forecaster.predict(steps=10, exog=df_exog_predict)

    expected = pd.Series(
                   data = np.array([2, 2, 2, 2, 2, 2, 2, 2, 2, 2]),
                   index = pd.RangeIndex(start=50, stop=60, step=1),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


def test_predict_output_when_categorical_features_native_implementation_LGBMClassifier_auto():
    """
    Test predict output when using LGBMClassifier and categorical variables with 
    categorical_features='auto'.
    """
    df_exog = pd.DataFrame({'exog_1': exog.to_numpy(),
                            'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
                            'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)})
    
    df_exog_predict = df_exog.iloc[:10, :].copy()
    df_exog_predict.index = pd.RangeIndex(start=50, stop=60)

    pipeline_categorical = make_pipeline(
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               FunctionTransformer(
                                   func=lambda x: x.astype('category'),
                                   feature_names_out= 'one-to-one'
                               )
                           )
    transformer_exog = make_column_transformer(
                            (
                                pipeline_categorical,
                                make_column_selector(dtype_exclude=np.number)
                            ),
                            remainder="passthrough",
                            verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterRecursiveClassifier(
                     estimator        = LGBMClassifier(verbose=-1, random_state=123),
                     lags             = 5,
                     transformer_exog = transformer_exog,
                     fit_kwargs       = {'categorical_feature': 'auto'}
                 )
    forecaster.fit(y=y, exog=df_exog)
    predictions = forecaster.predict(steps=10, exog=df_exog_predict)

    expected = pd.Series(
                   data = np.array([2, 1, 2, 2, 1, 1, 2, 2, 1, 3]),
                   index = pd.RangeIndex(start=50, stop=60, step=1),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)


@pytest.mark.parametrize("steps", 
                         [10, '2020-02-29', pd.to_datetime('2020-02-29')], 
                         ids=lambda steps: f'steps: {steps}')
def test_predict_output_when_window_features(steps):
    """
    Test output of predict when estimator is LGBMClassifier and window features.
    """
    
    rolling = RollingFeaturesClassification(stats=['proportion', 'entropy'], window_sizes=[3, 5])
    forecaster = ForecasterRecursiveClassifier(
        LGBMClassifier(verbose=-1, random_state=123), lags=3, window_features=rolling
    )
    forecaster.fit(y=y_dt, exog=exog_dt)
    predictions = forecaster.predict(steps=steps, exog=exog_dt_predict)

    expected = pd.Series(
                   data = np.array([3, 2, 2, 2, 1, 3, 2, 2, 1, 1]),
                   index = pd.date_range(start='2020-02-20', periods=10, freq='D'),
                   name = 'pred'
               )
    
    pd.testing.assert_series_equal(predictions, expected)



# Unit test show_versions
# ==============================================================================
from skforecast.utils import show_versions


def test_show_versions():
    """
    Test show versions function.
    """
    
    show_versions()
    s = show_versions(as_str=True)

    assert isinstance(s, str)
    assert "System" in s
    assert "Python dependencies" in s



# Unit test expand_index
# ==============================================================================
import re
import pytest
import pandas as pd
from skforecast.utils import expand_index


def test_TypeError_expand_index_when_steps_is_not_int():
    """
    Test TypeError is raised when `steps` is not a int.
    """
    index = pd.date_range(start='1990-01-01', periods=3, freq='D')
    steps = 2.5
    
    err_msg = re.escape(f"`steps` must be an integer. Got {type(steps)}.")
    with pytest.raises(TypeError, match=err_msg):
        expand_index(index, steps=steps)


def test_TypeError_expand_index_when_index_is_no_pandas_DatetimeIndex_or_RangeIndex():
    """
    Test TypeError is raised when input is not a pandas DatetimeIndex or RangeIndex.
    """
    index = pd.Index([0, 1, 2])

    err_msg = "Argument `index` must be a pandas DatetimeIndex or RangeIndex."
    with pytest.raises(TypeError, match = err_msg):
        expand_index(index, steps=3)


def test_output_expand_index_when_index_is_DatetimeIndex():
    """
    Test values returned by expand_index when input is DatetimeIndex.
    """
    index = pd.DatetimeIndex(['1990-01-01', '1990-01-02', '1990-01-03'],
                             dtype='datetime64[ns]', freq='D')
    expected = pd.DatetimeIndex(['1990-01-04', '1990-01-05', '1990-01-06'],
                             dtype='datetime64[ns]', freq='D')
    results = expand_index(index, steps=3)
    
    pd.testing.assert_index_equal(results, expected)


def test_output_expand_index_when_index_is_RangeIndex():
    """
    Test values returned by expand_index when input is RangeIndex.
    """
    index = pd.RangeIndex(start=0, stop=3, step=1)
    expected = pd.RangeIndex(start=3, stop=6, step=1)
    results  = expand_index(index, steps=3)
    
    pd.testing.assert_index_equal(results, expected)


def test_output_expand_index_when_index_is_not_pandas_index():
    """
    Test values returned by expand_index when input is not a pandas index.
    """
    index = ['0', '1', '2']
    expected = pd.RangeIndex(start=0, stop=3, step=1)
    results = expand_index(index, steps=3)
    
    pd.testing.assert_index_equal(results, expected)



# Unit test _binning_in_sample_residuals ForecasterRecursive
# ==============================================================================
import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression
from skforecast.recursive import ForecasterRecursive


def test_binning_in_sample_residuals_stored():
    """
    Test that _binning_in_sample_residuals stores residuals and binned residuals.
    """

    forecaster = ForecasterRecursive(
        estimator=object(), lags = 5, binner_kwargs={'n_bins': 3}
    )

    rng = np.random.default_rng(12345)
    y_pred = rng.normal(100, 15, 20)
    y_true = rng.normal(100, 10, 20)

    forecaster._binning_in_sample_residuals(
        y_pred=y_pred,
        y_true=y_true,
        store_in_sample_residuals=True
    )

    expected_residuals = np.array([
        34.58035615,  22.08911948,  15.6081091 ,   7.0808798 ,
        55.47454021,   1.80092458,  12.2624188 , -12.32822882,
        -0.451743  ,  11.83152763,  -7.11862253,   6.32581108,
        -20.92461257, -21.95291202, -10.55026794, -27.43753138,
        -6.24112163, -25.62685698,  -4.31327121, -18.40910724
    ])

    expected_residuals_binned = {
        0: np.array([
               34.58035615, 22.08911948, 15.6081091,  7.0808798, 55.47454021,
               1.80092458, 12.2624188
           ]),
        1: np.array([
               -12.32822882,  -0.45174300,  11.83152763,  -7.11862253,
               6.32581108, -20.92461257
           ]),
        2: np.array([
               -21.95291202, -10.55026794, -27.43753138,  -6.24112163,
               -25.62685698,  -4.31327121, -18.40910724
           ])
    }

    expected_binner_intervals = {
        0: (70.70705405481715, 90.25638761254116),
        1: (90.25638761254116, 109.36821559391004),
        2: (109.36821559391004, 135.2111448156828)
    }

    np.testing.assert_almost_equal(np.sort(forecaster.in_sample_residuals_), np.sort(expected_residuals))
    for k in expected_residuals_binned.keys():
        np.testing.assert_almost_equal(forecaster.in_sample_residuals_by_bin_[k], expected_residuals_binned[k])
    assert forecaster.binner_intervals_ == expected_binner_intervals


def test_binning_in_sample_residuals_store_in_sample_residuals_False():
    """
    Test that _binning_in_sample_residuals store_in_sample_residuals False.
    """

    forecaster = ForecasterRecursive(
        estimator=object(), lags = 5, binner_kwargs={'n_bins': 3}
    )

    rng = np.random.default_rng(12345)
    y_pred = rng.normal(100, 15, 20)
    y_true = rng.normal(100, 10, 20)

    forecaster._binning_in_sample_residuals(
        y_pred=y_pred,
        y_true=y_true,
        store_in_sample_residuals=False
    )
    
    expected_binner_intervals = {
        0: (70.70705405481715, 90.25638761254116),
        1: (90.25638761254116, 109.36821559391004),
        2: (109.36821559391004, 135.2111448156828)
    }

    assert forecaster.in_sample_residuals_ is None
    assert forecaster.in_sample_residuals_by_bin_ is None
    assert forecaster.binner_intervals_ == expected_binner_intervals


def test_binning_in_sample_residuals_probabilistic_mode_no_binned():
    """
    Test that _binning_in_sample_residuals stores when _probabilistic_mode is
    no binned.
    """

    forecaster = ForecasterRecursive(
        estimator=object(), lags = 5, binner_kwargs={'n_bins': 3}
    )

    rng = np.random.default_rng(12345)
    y_pred = rng.normal(100, 15, 20)
    y_true = rng.normal(100, 10, 20)

    forecaster._probabilistic_mode = "no_binned"
    forecaster._binning_in_sample_residuals(
        y_pred=y_pred,
        y_true=y_true,
        store_in_sample_residuals=True
    )
    
    expected_residuals = np.array([
        34.58035615,  22.08911948,  15.6081091 ,   7.0808798 ,
        55.47454021,   1.80092458,  12.2624188 , -12.32822882,
        -0.451743  ,  11.83152763,  -7.11862253,   6.32581108,
        -20.92461257, -21.95291202, -10.55026794, -27.43753138,
        -6.24112163, -25.62685698,  -4.31327121, -18.40910724
    ])

    np.testing.assert_almost_equal(np.sort(forecaster.in_sample_residuals_), np.sort(expected_residuals))
    assert forecaster.in_sample_residuals_by_bin_ is None
    assert forecaster.binner_intervals_ is None


def test_binning_in_sample_residuals_stores_maximum_10000_residuals():
    """
    Test that maximum 10_000 residuals are stored.
    """

    n = 15000
    y = pd.Series(
            data = np.random.normal(loc=10, scale=1, size=n),
            index = pd.date_range(start='01-01-2000', periods=n, freq='h')
        )
    forecaster = ForecasterRecursive(
        estimator=LinearRegression(), lags = 5, binner_kwargs={'n_bins': 2}
    )
    forecaster.fit(y, store_in_sample_residuals=True)
    max_residuals_per_bin = int(10_000 // forecaster.binner.n_bins_)

    assert len(forecaster.in_sample_residuals_) == 10_000
    for v in forecaster.in_sample_residuals_by_bin_.values():
        assert len(v) == max_residuals_per_bin



# Unit test set_lags ForecasterDirect
# ==============================================================================
import re
import pytest
import numpy as np
from sklearn.linear_model import LinearRegression
from skforecast.preprocessing import RollingFeatures
from skforecast.direct import ForecasterDirect


def test_set_lags_ValueError_when_lags_set_to_None_and_window_features_is_None():
    """
    Test ValueError is raised when lags is set to None and window_features is None.
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), steps=2, lags=3, window_features=None
    )

    err_msg = re.escape(
        "At least one of the arguments `lags` or `window_features` "
        "must be different from None. This is required to create the "
        "predictors used in training the forecaster."
    )
    with pytest.raises(ValueError, match = err_msg):
        forecaster.set_lags(lags=None)


@pytest.mark.parametrize("lags", 
                         [3, [1, 2, 3], np.array([1, 2, 3]), range(1, 4)],
                         ids = lambda lags: f'lags: {lags}')
def test_set_lags_with_different_inputs(lags):
    """
    Test how lags and max_lag attributes change with lags argument of different types.
    """
    forecaster = ForecasterDirect(
        estimator=LinearRegression(), lags=5, steps=2
    )
    forecaster.set_lags(lags=lags)

    np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2, 3]))
    assert forecaster.lags_names == ['lag_1', 'lag_2', 'lag_3']
    assert forecaster.max_lag == 3
    assert forecaster.window_size == 3


def test_set_lags_when_differentiation_is_not_None():
    """
    Test how `window_size` is also updated when the forecaster includes 
    differentiation.
    """
    forecaster = ForecasterDirect(
                     estimator       = LinearRegression(),
                     lags            = 3,
                     steps           = 2,
                     differentiation = 1
                 )
    
    np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2, 3]))
    assert forecaster.lags_names == ['lag_1', 'lag_2', 'lag_3']
    assert forecaster.max_lag == 3
    assert forecaster.window_size == 3 + 1
    assert forecaster.differentiator.window_size == 3 + 1

    forecaster.set_lags(lags=5)

    np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2, 3, 4, 5]))
    assert forecaster.lags_names == ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
    assert forecaster.max_lag == 5
    assert forecaster.window_size == 5 + 1
    assert forecaster.differentiator.window_size == 5 + 1


def test_set_lags_when_window_features():
    """
    Test how `window_size` is also updated when the forecaster includes
    window_features.
    """
    rolling = RollingFeatures(stats='mean', window_sizes=6)
    forecaster = ForecasterDirect(
                     estimator       = LinearRegression(),
                     lags            = 9,
                     steps           = 2,
                     window_features = rolling
                 )
    
    forecaster.set_lags(lags=5)

    np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2, 3, 4, 5]))
    assert forecaster.lags_names == ['lag_1', 'lag_2', 'lag_3', 'lag_4', 'lag_5']
    assert forecaster.max_lag == 5
    assert forecaster.max_size_window_features == 6
    assert forecaster.window_size == 6


@pytest.mark.parametrize("lags", 
                         [None, []],
                         ids = lambda lags: f'lags: {lags}')
def test_set_lags_to_None(lags):
    """
    Test how lags and max_lag attributes change when lags is set to None.
    """
    rolling = RollingFeatures(stats='mean', window_sizes=3)
    forecaster = ForecasterDirect(
                     estimator       = LinearRegression(),
                     lags            = 5,
                     steps           = 3,
                     window_features = rolling
                 )
    
    forecaster.set_lags(lags=lags)

    assert forecaster.lags is None
    assert forecaster.lags_names is None
    assert forecaster.max_lag is None
    assert forecaster.max_size_window_features == 3
    assert forecaster.window_size == 3



# Unit test _train_test_split_one_step_ahead ForecasterRecursiveClassifier
# ==============================================================================
import numpy as np
import pandas as pd
from sklearn.linear_model import LogisticRegression
from skforecast.recursive import ForecasterRecursiveClassifier


def test_train_test_split_one_step_ahead_when_y_is_series_15_and_exog_is_dataframe_of_float_int():
    """
    Test the output of _train_test_split_one_step_ahead when 
    y=pd.Series(np.arange(10)) and exog is a pandas dataframe with 
    2 columns of floats or ints.
    """
    y = pd.Series(
        np.array(['a', 'b', 'c', 'a', 'b', 'c', 'a', 'b', 'c', 'a', 'b', 'c', 'a', 'b', 'c']), 
        index=pd.date_range("2020-01-01", periods=15),
        name='y'
    )
    exog = pd.DataFrame(
        {
            "exog_1": np.arange(100, 115, dtype=float),
            "exog_2": np.arange(1000, 1015, dtype=float),
        },
        index=pd.date_range("2020-01-01", periods=15),
    )

    forecaster = ForecasterRecursiveClassifier(LogisticRegression(), lags=5)

    X_train, y_train, X_test, y_test = forecaster._train_test_split_one_step_ahead(
        y=y, exog=exog, initial_train_size=10
    )

    expected_X_train = pd.DataFrame(
        {
            "lag_1": [1., 2., 0., 1., 2.],
            "lag_2": [0., 1., 2., 0., 1.],
            "lag_3": [2., 0., 1., 2., 0.],
            "lag_4": [1., 2., 0., 1., 2.],
            "lag_5": [0., 1., 2., 0., 1.],
            "exog_1": [105., 106., 107., 108., 109.],
            "exog_2": [1005., 1006., 1007., 1008., 1009.],
        },
        index=pd.DatetimeIndex(
            ["2020-01-06", "2020-01-07", "2020-01-08", "2020-01-09", "2020-01-10"],
            freq='D'
        ),
    )
    expected_y_train = pd.Series(
        [2., 0., 1., 2., 0.],
        index=pd.DatetimeIndex(
            ["2020-01-06", "2020-01-07", "2020-01-08", "2020-01-09", "2020-01-10"],
            freq='D'
        ),
        name = 'y'
    )

    expected_X_test = pd.DataFrame(
        {
            "lag_1": [0., 1., 2., 0., 1.],
            "lag_2": [2., 0., 1., 2., 0.],
            "lag_3": [1., 2., 0., 1., 2.],
            "lag_4": [0., 1., 2., 0., 1.],
            "lag_5": [2., 0., 1., 2., 0.],
            "exog_1": [110., 111., 112., 113., 114.],
            "exog_2": [1010., 1011., 1012., 1013., 1014.],
        },
        index=pd.DatetimeIndex(
            ["2020-01-11", "2020-01-12", "2020-01-13", "2020-01-14", "2020-01-15"],
            freq='D'
        ),
    )
    expected_y_test = pd.Series(
        [1., 2., 0., 1., 2.],
        index=pd.DatetimeIndex(
            ["2020-01-11", "2020-01-12", "2020-01-13", "2020-01-14", "2020-01-15"],
            freq='D'
        ),
        name = 'y'
    )

    pd.testing.assert_frame_equal(X_train, expected_X_train)
    pd.testing.assert_series_equal(y_train, expected_y_train)
    pd.testing.assert_frame_equal(X_test, expected_X_test)
    pd.testing.assert_series_equal(y_test, expected_y_test)



# Unit test _create_predict_inputs ForecasterRecursive
# ==============================================================================
import re
import pytest
import numpy as np
import pandas as pd
from sklearn.exceptions import NotFittedError
from skforecast.recursive import ForecasterRecursive
from sklearn.compose import ColumnTransformer
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import OneHotEncoder
from sklearn.preprocessing import OrdinalEncoder
from sklearn.compose import make_column_transformer
from sklearn.linear_model import LinearRegression
from sklearn.ensemble import HistGradientBoostingRegressor

# Fixtures
from .fixtures_forecaster_recursive import y as y_categorical
from .fixtures_forecaster_recursive import exog as exog_categorical
from .fixtures_forecaster_recursive import data  # to test results when using differentiation


def test_create_predict_inputs_NotFittedError_when_fitted_is_False():
    """
    Test NotFittedError is raised when fitted is False.
    """
    forecaster = ForecasterRecursive(
                     estimator = LinearRegression(),
                     lags      = 5
                 )

    err_msg = re.escape(
        "This Forecaster instance is not fitted yet. Call `fit` with "
        "appropriate arguments before using predict."
    )
    with pytest.raises(NotFittedError, match = err_msg):
        forecaster._create_predict_inputs(steps=5)


def test_create_predict_inputs_when_estimator_is_LinearRegression():
    """
    Test _create_predict_inputs when using LinearRegression as estimator.
    """
    forecaster = ForecasterRecursive(
                     estimator = LinearRegression(),
                     lags      = 5
                 )
    forecaster.fit(y=pd.Series(np.arange(50, dtype=float), name='y'))
    results = forecaster._create_predict_inputs(steps=5)

    expected = (
        np.array([45., 46., 47., 48., 49.]),
        None,
        pd.RangeIndex(start=50, stop=55, step=1),
        5
    )
    
    np.testing.assert_array_almost_equal(results[0], expected[0])
    assert results[1] is None
    pd.testing.assert_index_equal(results[2], expected[2])
    assert results[3] == expected[3]


def test_create_predict_inputs_when_with_transform_y():
    """
    Test _create_predict_inputs when using LinearRegression as estimator and StandardScaler.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9,  1.09, -3.61,  0.72, -0.11, -0.4,  0.49,
                       0.67,  0.54, -0.17,  0.54,  1.49, -2.26, -0.41, -0.64, -0.8,
                      -0.61, -0.88]),
            name = 'y'
        )

    forecaster = ForecasterRecursive(
                     estimator     = LinearRegression(),
                     lags          = 5,
                     transformer_y = StandardScaler()
                 )
    forecaster.fit(y=y)
    results = forecaster._create_predict_inputs(steps=5)

    expected = (
        np.array([-0.1056608, -0.30987914, -0.45194408, -0.28324197, -0.52297655]),
        None,
        pd.RangeIndex(start=20, stop=25, step=1),
        5
    )
    
    np.testing.assert_array_almost_equal(results[0], expected[0])
    assert results[1] is None
    pd.testing.assert_index_equal(results[2], expected[2])
    assert results[3] == expected[3]


def test_create_predict_inputs_when_with_transform_y_and_transform_exog_series():
    """
    Test _create_predict_inputs when using LinearRegression as estimator, StandardScaler
    as transformer_y and StandardScaler as transformer_exog.
    """
    y = pd.Series(np.array([-0.59,  0.02, -0.9,  1.09, -3.61,  0.72, -0.11, -0.4]))
    exog = pd.Series(np.array([7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4]), name='exog')
    exog_predict = exog.copy()
    exog_predict.index = pd.RangeIndex(start=8, stop=16)

    forecaster = ForecasterRecursive(
                     estimator        = LinearRegression(),
                     lags             = 5,
                     transformer_y    = StandardScaler(),
                     transformer_exog = StandardScaler()
                 )
    forecaster.fit(y=y, exog=exog)
    results = forecaster._create_predict_inputs(steps=5, exog=exog_predict)

    expected = (
        np.array([1.16937289, -2.34810076, 0.89246539, 0.27129451, 0.0542589]),
        np.array([[-1.76425513], [-1.00989936], [0.59254869], [0.45863938], [0.1640389]]),
        pd.RangeIndex(start=8, stop=13, step=1),
        5
    )
    
    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])
    pd.testing.assert_index_equal(results[2], expected[2])
    assert results[3] == expected[3]


def test_create_predict_inputs_when_with_transform_y_and_transform_exog_df():
    """
    Test _create_predict_inputs when using LinearRegression as estimator, StandardScaler
    as transformer_y and transformer_exog as transformer_exog.
    """
    y = pd.Series(
            np.array([-0.59,  0.02, -0.9 ,  1.09, -3.61,  0.72, -0.11, -0.4])
        )
    exog = pd.DataFrame({
               'col_1': [7.5, 24.4, 60.3, 57.3, 50.7, 41.4, 87.2, 47.4],
               'col_2': ['a', 'a', 'a', 'a', 'b', 'b', 'b', 'b']}
           )
    exog_predict = exog.copy()
    exog_predict.index = pd.RangeIndex(start=8, stop=16)

    transformer_y = StandardScaler()
    transformer_exog = ColumnTransformer(
                            [('scale', StandardScaler(), ['col_1']),
                             ('onehot', OneHotEncoder(), ['col_2'])],
                            remainder = 'passthrough',
                            verbose_feature_names_out = False
                       )
    
    forecaster = ForecasterRecursive(
                     estimator        = LinearRegression(),
                     lags             = 5,
                     transformer_y    = transformer_y,
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y, exog=exog)
    results = forecaster._create_predict_inputs(steps=5, exog=exog_predict)
    
    expected = (
        np.array([1.16937289, -2.34810076, 0.89246539,  0.27129451,  0.0542589]),
        np.array([[-1.76425513,  1.        ,  0.        ],
                  [-1.00989936,  1.        ,  0.        ],
                  [ 0.59254869,  1.        ,  0.        ],
                  [ 0.45863938,  1.        ,  0.        ],
                  [ 0.1640389 ,  0.        ,  1.        ]]),
        pd.RangeIndex(start=8, stop=13, step=1),
        5
    )
    
    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])
    pd.testing.assert_index_equal(results[2], expected[2])
    assert results[3] == expected[3]


def test_create_predict_inputs_when_categorical_features_native_implementation_HistGradientBoostingRegressor():
    """
    Test _create_predict_inputs when using HistGradientBoostingRegressor and categorical variables.
    """
    df_exog = pd.DataFrame(
        {'exog_1': exog_categorical,
         'exog_2': ['a', 'b', 'c', 'd', 'e'] * 10,
         'exog_3': pd.Categorical(['F', 'G', 'H', 'I', 'J'] * 10)}
    )
    
    exog_predict = df_exog.copy()
    exog_predict.index = pd.RangeIndex(start=50, stop=100)

    categorical_features = df_exog.select_dtypes(exclude=[np.number]).columns.tolist()
    transformer_exog = make_column_transformer(
                           (
                               OrdinalEncoder(
                                   dtype=int,
                                   handle_unknown="use_encoded_value",
                                   unknown_value=-1,
                                   encoded_missing_value=-1
                               ),
                               categorical_features
                           ),
                           remainder="passthrough",
                           verbose_feature_names_out=False,
                       ).set_output(transform="pandas")
    
    forecaster = ForecasterRecursive(
                     estimator        = HistGradientBoostingRegressor(
                                            categorical_features = categorical_features,
                                            random_state         = 123
                                        ),
                     lags             = 5,
                     transformer_y    = None,
                     transformer_exog = transformer_exog
                 )
    forecaster.fit(y=y_categorical, exog=df_exog)
    results = forecaster._create_predict_inputs(steps=10, exog=exog_predict)
    
    expected = (
        np.array([0.25045537, 0.48303426, 0.98555979, 0.51948512, 0.61289453]),
        np.array([[0.        , 0.        , 0.12062867],
                  [1.        , 1.        , 0.8263408 ],
                  [2.        , 2.        , 0.60306013],
                  [3.        , 3.        , 0.54506801],
                  [4.        , 4.        , 0.34276383],
                  [0.        , 0.        , 0.30412079],
                  [1.        , 1.        , 0.41702221],
                  [2.        , 2.        , 0.68130077],
                  [3.        , 3.        , 0.87545684],
                  [4.        , 4.        , 0.51042234]]),
        pd.RangeIndex(start=50, stop=60, step=1),
        10
    )
    
    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])
    pd.testing.assert_index_equal(results[2], expected[2])
    assert results[3] == expected[3]


def test_create_predict_inputs_when_with_exog_differentiation_is_1():
    """
    Test _create_predict_inputs when using LinearRegression as estimator 
    and differentiation=1.
    """

    end_train = '2003-03-01 23:59:00'

    # Simulated exogenous variable
    rng = np.random.default_rng(9876)
    exog = pd.Series(
        rng.normal(loc=0, scale=1, size=len(data)), index=data.index, name='exog'
    )
    steps = data.index[-1]

    forecaster = ForecasterRecursive(
                     estimator       = LinearRegression(),
                     lags            = 15,
                     differentiation = 1
                )
    forecaster.fit(y=data.loc[:end_train], exog=exog.loc[:end_train])
    results = forecaster._create_predict_inputs(steps=steps, exog=exog.loc[end_train:])
    
    expected = (
        np.array(
            [np.nan, 0.14355438, -0.56028323,  0.07558021, 0.04869748,
             0.09807633, -0.00584249,  0.17596768,  0.01630394,  0.09883014,
             0.02377842, -0.01018012,  0.10597971, -0.01463081, -0.48984868,
             0.07503713]
        ),
        np.array([[ 1.16172882], [ 0.29468848], [-0.4399757 ], [ 1.25008389],
                  [ 1.37496887], [-0.41673182], [ 0.32732157], [ 1.57848827],
                  [ 0.40402941], [ 1.34466867], [-1.0724777 ], [ 0.14619469],
                  [-0.78475177], [-2.97987806], [-1.46735738], [ 0.84295053],
                  [ 0.37485   ], [ 0.03500746], [-1.37307705], [ 1.43220391],
                  [ 0.76682773], [ 0.50114842], [ 0.839902  ], [-1.0012572 ],
                  [-1.67885896], [-0.81587204], [-1.33508966], [-0.42014975], 
                  [ 0.98579761], [-0.17262001], [-0.02161134], [-0.62664312],
                  [-0.91997057], [-1.14759504], [ 0.18661806], [ 0.15897422], 
                  [-3.20047182], [-0.5236335 ], [-1.00620967], [ 0.12763016], 
                  [-0.54508051], [ 0.13153937], [-0.14105938], [-0.16935134], 
                  [ 0.06279071], [-0.55075912], [-0.22425219], [ 2.00991015], 
                  [ 0.79230622], [-0.55605221], [-0.27088044], [-0.39179496], 
                  [-1.34347494], [-0.07230096], [-1.51146602], [-0.66840335], 
                  [-0.78148407], [-0.27354003], [-0.27128144], [-0.6389055 ], 
                  [ 0.19573233], [-0.67321672], [ 1.1559056 ]]
        ),
        pd.date_range(start='2003-04-01', periods=63, freq='MS'),
        63
    )
    
    np.testing.assert_array_almost_equal(results[0], expected[0])
    np.testing.assert_array_almost_equal(results[1], expected[1])
    pd.testing.assert_index_equal(results[2], expected[2])
    assert results[3] == expected[3]



# Unit test __init__ ForecasterRecursiveClassifier
# ==============================================================================
import re
import pytest
import numpy as np
from lightgbm import LGBMClassifier
from sklearn.linear_model import LogisticRegression
from skforecast.preprocessing import RollingFeaturesClassification
from skforecast.recursive import ForecasterRecursiveClassifier


def test_init_ValueError_when_features_encoding_not_valid():
    """
    Test ValueError is raised when `features_encoding` is not valid.
    """
    err_msg = re.escape(
        "`features_encoding` must be one of ['auto', 'categorical', 'ordinal']. "
        "Got 'not_valid'."
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterRecursiveClassifier(
            estimator         = LogisticRegression(),
            lags              = 3,
            features_encoding = 'not_valid'
        )


def test_init_ValueError_when_estimator_does_not_support_categorical_features():
    """
    Test ValueError is raised when `features_encoding='categorical'` and
    the estimator does not support categorical features natively.
    """
    err_msg = re.escape(
        f"`features_encoding='categorical'` requires a estimator that "
        f"supports native categorical features (LightGBM, CatBoost, XGBoost). "
        f"Got {type(LogisticRegression()).__name__}. Use 'auto' or 'ordinal' instead."
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterRecursiveClassifier(
            estimator         = LogisticRegression(),
            lags              = 3,
            features_encoding = 'categorical'
        )


def test_init_ValueError_when_no_lags_or_window_features():
    """
    Test ValueError is raised when no lags or window_features are passed.
    """
    err_msg = re.escape(
        "At least one of the arguments `lags` or `window_features` "
        "must be different from None. This is required to create the "
        "predictors used in training the forecaster."
    )
    with pytest.raises(ValueError, match = err_msg):
        ForecasterRecursiveClassifier(
            estimator       = LogisticRegression(),
            lags            = None,
            window_features = None
        )


@pytest.mark.parametrize("features_encoding, estimator, expected", 
                         [('auto', LogisticRegression(), False), 
                          ('auto', LGBMClassifier(verbose=-1), True), 
                          ('categorical', LGBMClassifier(verbose=-1), True), 
                          ('ordinal', LGBMClassifier(verbose=-1), False)], 
                         ids = lambda dt: f'features_encoding, estimator, expected: {dt}')
def test_init_use_native_categoricals_set(features_encoding, estimator, expected):
    """
    Test use_native_categoricals is correctly set during initialization.
    """

    forecaster = ForecasterRecursiveClassifier(
                     estimator         = estimator,
                     lags              = 3,
                     features_encoding = features_encoding
                 )
    
    assert forecaster.use_native_categoricals == expected


@pytest.mark.parametrize("lags, window_features, expected", 
                         [(5, None, 5), 
                          (None, True, 6), 
                          ([], True, 6), 
                          (5, True, 6)], 
                         ids = lambda dt: f'lags, window_features, expected: {dt}')
def test_init_window_size_correctly_stored(lags, window_features, expected):
    """
    Test window_size is correctly stored when lags or window_features are passed.
    """
    if window_features:
        window_features = RollingFeaturesClassification(
            stats=['proportion', 'mode'], window_sizes=[5, 6]
        )

    forecaster = ForecasterRecursiveClassifier(
                     estimator       = LogisticRegression(),
                     lags            = lags,
                     window_features = window_features
                 )
    
    assert forecaster.window_size == expected
    if lags:
        np.testing.assert_array_almost_equal(forecaster.lags, np.array([1, 2, 3, 4, 5]))
        assert forecaster.lags_names == [f'lag_{i}' for i in range(1, lags + 1)]
        assert forecaster.max_lag == lags
    else:
        assert forecaster.lags is None
        assert forecaster.lags_names is None
        assert forecaster.max_lag is None
    if window_features:
        assert forecaster.window_features_names == ['roll_proportion_5', 'roll_mode_6']
        assert forecaster.window_features_class_names == ['RollingFeaturesClassification']
    else:
        assert forecaster.window_features_names is None
        assert forecaster.window_features_class_names is None
